# -*- coding: utf-8 -*-

import numpy as np


def singleton(cls):
    instances = {}

    def getinstance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return getinstance


@singleton
class DefaultRandomState(np.random.RandomState):
    def __init__(self, seed=None):
        super().__init__(seed)



# -*- coding: utf-8 -*-

import math

import torch
from torch import nn
from torch.nn import functional

from jack.util.torch import misc


class ConvCharEmbeddingModule(nn.Module):
    def __init__(self, num_chars, size, conv_width=5):
        super(ConvCharEmbeddingModule, self).__init__()
        self._size = size
        self._conv_width = conv_width
        self._embeddings = torch.nn.Embedding(num_chars, size)
        self._embeddings.weight.data.mul_(0.1)
        self._conv = torch.nn.Conv1d(size, size, conv_width, padding=math.floor(conv_width / 2))

    def forward(self, unique_word_chars, unique_word_lengths, sequences_as_uniqs=None):
        long_tensor = torch.cuda.LongTensor if torch.cuda.device_count() > 0 else torch.LongTensor
        embedded_chars = self._embeddings(unique_word_chars.type(long_tensor))
        # [N, S, L]
        conv_out = self._conv(embedded_chars.transpose(1, 2))
        # [N, L]
        conv_mask = misc.mask_for_lengths(unique_word_lengths)
        conv_out = conv_out + conv_mask.unsqueeze(1)
        embedded_words = conv_out.max(2)[0]

        if sequences_as_uniqs is None:
            return embedded_words
        else:
            if not isinstance(sequences_as_uniqs, list):
                sequences_as_uniqs = [sequences_as_uniqs]

            all_embedded = []
            for word_idxs in sequences_as_uniqs:
                all_embedded.append(functional.embedding(
                    word_idxs.type(long_tensor), embedded_words))
            return all_embedded



# -*- coding: utf-8 -*-

import abc
import sys

import tensorflow as tf

from jack.readers.link_prediction.similarities import negative_l1_distance


class BaseModel(metaclass=abc.ABCMeta):
    def __init__(self, subject_embeddings=None, predicate_embeddings=None, object_embeddings=None, *args, **kwargs):
        """
        Abstract class inherited by all models.

        Args:
            subject_embeddings: (batch_size, entity_embedding_size) Tensor.
            predicate_embeddings: (batch_size, predicate_embedding_size) Tensor.
            object_embeddings: (batch_size, entity_embedding_size) Tensor.
        """
        self.subject_embeddings = subject_embeddings
        self.predicate_embeddings = predicate_embeddings
        self.object_embeddings = object_embeddings

    @abc.abstractmethod
    def __call__(self):
        raise NotImplementedError

    @property
    def parameters(self):
        return []


class TranslatingModel(BaseModel):
    def __init__(self, similarity_function=negative_l1_distance, *args, **kwargs):
        """
        Implementation of the Translating Embeddings model [1].
        [1] Bordes, A. et al. - Translating Embeddings for Modeling Multi-relational Data - NIPS 2013

        Args:
            similarity_function: Similarity function.
        """
        super().__init__(*args, **kwargs)
        self.similarity_function = similarity_function

    def __call__(self):
        """
        :return: (batch_size) Tensor containing the scores associated by the models to the walks.
        """
        translated_subject_embedding = self.subject_embeddings + self.predicate_embeddings
        return self.similarity_function(translated_subject_embedding, self.object_embeddings)


class BilinearDiagonalModel(BaseModel):
    def __init__(self, *args, **kwargs):
        """
        Implementation of the Bilinear-Diagonal model [1]
        [1] Yang, B. et al. - Embedding Entities and Relations for Learning and Inference in Knowledge Bases - ICLR 2015

        Args:
            similarity_function: Similarity function.
        """
        super().__init__(*args, **kwargs)

    def __call__(self):
        """
        :return: (batch_size) Tensor containing the scores associated by the models to the walks.
        """
        scaled_subject_embedding = self.subject_embeddings * self.predicate_embeddings
        return tf.reduce_sum(scaled_subject_embedding * self.object_embeddings, axis=1)


class BilinearModel(BaseModel):
    def __init__(self, *args, **kwargs):
        """
        Implementation of the Bilinear model [1]
        [1] Nickel, M. et al. - A Three-Way Model for Collective Learning on Multi-Relational Data - ICML 2011

        ArgS:
            similarity_function: Similarity function.
        """
        super().__init__(*args, **kwargs)

    def __call__(self):
        """
        :return: (batch_size) Tensor containing the scores associated by the models to the walks.
        """
        es, emb_size = tf.expand_dims(self.subject_embeddings, 1), tf.shape(self.subject_embeddings)[1]
        W = tf.reshape(self.predicate_embeddings, (-1, emb_size, emb_size))
        sW = tf.matmul(es, W)[:, 0, :]
        return tf.reduce_sum(sW * self.object_embeddings, axis=1)


class ComplexModel(BaseModel):
    def __init__(self, *args, **kwargs):
        """
        Implementation of the ComplEx model [1]
        [1] Trouillon, T. et al. - Complex Embeddings for Simple Link Prediction - ICML 2016

        Args:
            embedding size: Embedding size.
        """
        super().__init__(*args, **kwargs)

    def __call__(self):
        """
        :return: (batch_size) Tensor containing the scores associated by the models to the walks.
        """
        es_re, es_im = tf.split(value=self.subject_embeddings, num_or_size_splits=2, axis=1)
        eo_re, eo_im = tf.split(value=self.object_embeddings, num_or_size_splits=2, axis=1)
        ew_re, ew_im = tf.split(value=self.predicate_embeddings, num_or_size_splits=2, axis=1)

        def dot3(arg1, rel, arg2):
            return tf.reduce_sum(arg1 * rel * arg2, axis=1)

        score = dot3(es_re, ew_re, eo_re) + dot3(es_re, ew_im, eo_im) + dot3(es_im, ew_re, eo_im) - dot3(es_im, ew_im, eo_re)
        return score

# Aliases
TransE = TranslatingEmbeddings = TranslatingModel
DistMult = BilinearDiagonal = BilinearDiagonalModel
RESCAL = Bilinear = BilinearModel
ComplEx = ComplexE = ComplexModel


def get_function(function_name):
    this_module = sys.modules[__name__]
    if not hasattr(this_module, function_name):
        raise ValueError('Unknown model: {}'.format(function_name))
    return getattr(this_module, function_name)



"""
This files allows creating a jtr datafile for the SNLI corpus,
whereby each instance receives support under the form of 'related' instances
"""


import json

__candidate_labels = ['entailment', 'neutral', 'contradiction']
__candidates = [{'text': cl} for cl in __candidate_labels]


def convert_snli(multisnli_file):
    """ io SNLI files into jack format.
    Data source: http://nlp.stanford.edu/projects/snli/snli_1.0.zip
    Files to be converted: snli_1.0_dev.jsonl, snli_1.0_train.jsonl, snli_1.0_test.jsonl
    (the *.txt files contain the same data in a different format)

    Format:
        - support = the premise = 'sentence1' in original SNLI data (as id, use 'captionID', the id of sentence1)
        - question = the hypothesis = 'sentence2' in original SNLI data
    Notes:
        - instances with gold labels '-' are removed from the corpus
    """
    assert 'multinli_0.9' in multisnli_file and multisnli_file.endswith('.txt'),\
        "input should be the multinli_0.9_X.txt files (X=test/train/dev)"

    with open(multisnli_file, 'r') as f:
        data = [__convert_snli_instance(line.strip().strip("\n").split("\t")) for line in f.readlines()]

        return {'meta': 'MultiSNLI',
                'globals': {'candidates': __candidates},
                'instances': [d for d in data if d]  # filter out invalid ones
                }


def __convert_snli_instance(lspl):
    if len(lspl) != 15:
        return None
    try:
        gold_label, _, _, _, _, sentence1, sentence2, promptID, pairID, genre, _, _, _, _, _ = lspl
        if gold_label == "gold_label":
            return None

        if not gold_label in __candidate_labels:
            raise IOError('invalid gold label')
        queb = {}
        queb['id'] = pairID
        queb['genre'] = genre,
        queb['support'] = [
            {'id': promptID, 'text': sentence1}]
        queb['questions'] = [
            {'question': sentence2,
             'answers': [
                 # {'index': __candidate_labels.index(instance['gold_label'])},
                 {'text': __candidate_labels[__candidate_labels.index(gold_label)]}]}]

        return queb

    except IOError:
        return None


def main():
    import sys
    if len(sys.argv) == 2:
        corpus = convert_snli(sys.argv[1])
    else:
        for corpus_name in ["dev_matched", "train"]:
            corpus = convert_snli("../../data/MultiNLI/multinli_0.9_%s.txt" % corpus_name)
            with open("../../data/MultiNLI/multinli_0.9_%s_jtr.json" % corpus_name, 'w') as outfile:
                print("Create file snli_0.9_%s_jtr.txt" % corpus_name)
                json.dump(corpus, outfile, indent=2)

        # create train set test data
        corpus = convert_snli("../../data/MultiNLI/multinli_0.9_train.txt")
        corpus['instances'] = corpus['instances'][:2000]
        with open("../../tests/test_data/MultiNLI/2000_samples_train_jtr.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        corpus['instances'] = corpus['instances'][:100]
        with open("../../tests/test_data/MultiNLI/overfit.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        # create snippets and overfit test data
        corpus['instances'] = corpus['instances'][:10]
        with open("../../data/MultiNLI/multinli_0.9_debug_jtr.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)
        with open("../../data/MultiNLI/snippet.jack.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        # create dev set test data
        corpus = convert_snli("../../data/MultiNLI/multinli_0.9_dev_matched.txt")
        corpus['instances'] = corpus['instances'][:1000]
        with open("../../tests/test_data/MultiNLI/1000_samples_dev_jtr.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        # create dev set test data
        corpus = convert_snli("../../data/MultiNLI/multinli_0.9_train.txt")
        corpus['instances'] = corpus['instances'][:2000]
        with open("../../tests/test_data/MultiNLI/2000_samples_train_jtr.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
import math

import tensorflow as tf

from jack.util.tf import segment, misc


def attention_softmax(attn_scores, length=None):
    if length is not None:
        attn_scores += misc.mask_for_lengths(length, tf.shape(attn_scores)[2])
    return tf.nn.softmax(attn_scores)


def apply_attention(attn_scores, states, length, is_self=False, with_sentinel=True, reuse=False, seq2_to_seq1=None):
    attn_scores += tf.expand_dims(misc.mask_for_lengths(length, tf.shape(attn_scores)[2]), 1)
    softmax = tf.nn.softmax if seq2_to_seq1 is None else lambda x: segment.segment_softmax(x, seq2_to_seq1)
    if is_self:
        # exclude attending to state itself
        attn_scores += tf.expand_dims(tf.diag(tf.fill([tf.shape(attn_scores)[1]], -1e6)), 0)
    if with_sentinel:
        with tf.variable_scope('sentinel', reuse=reuse):
            s = tf.get_variable('score', [1, 1, 1], tf.float32, tf.zeros_initializer())
        s = tf.tile(s, [tf.shape(attn_scores)[0], tf.shape(attn_scores)[1], 1])
        attn_probs = softmax(tf.concat([s, attn_scores], 2))
        attn_probs = attn_probs[:, :, 1:]
    else:
        attn_probs = softmax(attn_scores)
    attn_states = tf.einsum('abd,adc->abc', attn_probs, states)
    if seq2_to_seq1 is not None:
        attn_states = tf.unsorted_segment_sum(attn_states, seq2_to_seq1, tf.reduce_max(seq2_to_seq1) + 1)
    return attn_scores, attn_probs, attn_states


def dot_attention(seq1, seq2, len2, scaled=True, with_sentinel=True, seq2_to_seq1=None, **kwargs):
    query, key, value = _get_query_key_value(seq1, seq2, **kwargs)
    if seq2_to_seq1 is not None:
        query = tf.gather(query, seq2_to_seq1)
    attn_scores = tf.einsum('abc,adc->abd', query, key)
    if scaled:
        attn_scores /= tf.sqrt(float(query.get_shape()[-1].value))
    return apply_attention(attn_scores, value, len2, seq1 is seq2, with_sentinel, seq2_to_seq1=seq2_to_seq1)


def bilinear_attention(seq1, seq2, len2, scaled=True, with_sentinel=True, seq2_to_seq1=None, **kwargs):
    query, key, value = _get_query_key_value(seq1, seq2, **kwargs)
    if seq2_to_seq1 is not None:
        query = tf.gather(query, seq2_to_seq1)
    attn_scores = tf.einsum('abc,adc->abd', tf.layers.dense(query, key.get_shape()[-1].value, use_bias=False), key)
    attn_scores += tf.layers.dense(query, 1, use_bias=False)
    attn_scores += tf.transpose(tf.layers.dense(key, 1, use_bias=False), [0, 2, 1])
    if scaled:
        attn_scores /= math.sqrt(float(query.get_shape()[-1].value))
    return apply_attention(attn_scores, value, len2, seq1 is seq2, with_sentinel, seq2_to_seq1=seq2_to_seq1)


def diagonal_bilinear_attention(seq1, seq2, len2, scaled=True, with_sentinel=True, seq2_to_seq1=None, **kwargs):
    query, key, value = _get_query_key_value(seq1, seq2, **kwargs)
    if seq2_to_seq1 is not None:
        query = tf.gather(query, seq2_to_seq1)
    v = tf.get_variable('attn_weight', [1, 1, query.get_shape()[-1].value], tf.float32,
                        initializer=tf.ones_initializer())
    attn_scores = tf.einsum('abc,adc->abd', v * query, key)
    attn_scores += tf.layers.dense(query, 1, use_bias=False)
    attn_scores += tf.transpose(tf.layers.dense(key, 1, use_bias=False), [0, 2, 1])
    if scaled:
        attn_scores /= math.sqrt(float(query.get_shape()[-1].value))
    return apply_attention(attn_scores, value, len2, seq1 is seq2, with_sentinel, seq2_to_seq1=seq2_to_seq1)


def mlp_attention(hidden_dim, seq1, seq2, len2, activation=tf.nn.relu, with_sentinel=True, seq2_to_seq1=None, **kwargs):
    query, key, value = _get_query_key_value(seq1, seq2, **kwargs)
    if seq2_to_seq1 is not None:
        query = tf.gather(query, seq2_to_seq1)
    hidden1 = tf.layers.dense(query, hidden_dim)
    hidden2 = tf.layers.dense(key, hidden_dim)
    hidden = tf.tile(tf.expand_dims(hidden1, 2), [1, 1, tf.shape(key)[1], -1]) + tf.expand_dims(hidden2, 1)
    attn_scores = tf.layers.dense(activation(hidden), 1, use_bias=with_sentinel)
    return apply_attention(tf.squeeze(attn_scores, 3), value, len2, seq1 is seq2, with_sentinel,
                           seq2_to_seq1=seq2_to_seq1)


def _get_query_key_value(seq1, seq2, key_value_attn=False, repr_dim=None, **kwargs):
    repr_dim = repr_dim or seq2.get_shape()[-1].value
    if key_value_attn:
        with tf.variable_scope('key_value_projection') as vs:
            key = tf.layers.dense(seq2, repr_dim, name='key')
            value = tf.layers.dense(seq2, repr_dim, name='value')
            query = tf.layers.dense(seq1, repr_dim, name='query')
        return query, key, value
    else:
        return seq1, seq2, seq2


def coattention(seq1, len1, seq2, len2, scaled=True, with_sentinel=True, attn_fn=dot_attention, seq2_to_seq1=None):
    attn_scores, attn_probs1, attn_states1 = attn_fn(seq1, seq2, len2, scaled, with_sentinel, seq2_to_seq1)
    _, _, attn_states2 = apply_attention(
        tf.transpose(attn_scores, [0, 2, 1]), seq1, len1, seq1 is seq2, with_sentinel, reuse=True)
    co_attn_state = tf.einsum('abd,adc->abc', attn_probs1, attn_states2)

    return attn_scores, attn_probs1, attn_states1, attn_states2, co_attn_state


def attention_softmax3d(values):
    """
    Performs a softmax over the attention values.
    Args:
        values: tensor with shape (batch_size, time_steps, time_steps)
    Returns:
        tensor with shape (batch_size, time_steps, time_steps)
    """
    original_shape = tf.shape(values)
    # tensor with shape (batch_size * time_steps, time_steps)
    reshaped_values = tf.reshape(tensor=values, shape=[-1, original_shape[2]])
    # tensor with shape (batch_size * time_steps, time_steps)
    softmax_reshaped_values = tf.nn.softmax(reshaped_values)
    # tensor with shape (batch_size, time_steps, time_steps)
    return tf.reshape(softmax_reshaped_values, original_shape)


def distance_biases(time_steps, window_size=10, reuse=False):
    """
    Return a 2-d tensor with the values of the distance biases to be applied
    on the intra-attention matrix of size sentence_size

    Args:
        time_steps: tensor scalar
        window_size: window size
        reuse: reuse variables
    Returns:
        2-d tensor (time_steps, time_steps)
    """
    with tf.variable_scope('distance-bias', reuse=reuse):
        # this is d_{i-j}
        distance_bias = tf.get_variable('dist_bias', [window_size], initializer=tf.zeros_initializer())
        r = tf.range(0, time_steps)
        r_matrix = tf.tile(tf.reshape(r, [1, -1]), tf.stack([time_steps, 1]))
        raw_idxs = r_matrix - tf.reshape(r, [-1, 1])
        clipped_idxs = tf.clip_by_value(raw_idxs, 0, window_size - 1)
        values = tf.nn.embedding_lookup(distance_bias, clipped_idxs)
    return values



# -*- coding: utf-8 -*-

import tensorflow as tf

from jack.util.tf.segment import segment_softmax


def xqa_crossentropy_loss(start_scores, end_scores, answer_span, answer2support, support2question, use_sum=True):
    """Very common XQA loss function."""
    num_questions = tf.reduce_max(support2question) + 1

    start, end = answer_span[:, 0], answer_span[:, 1]

    start_probs = segment_softmax(start_scores, support2question)
    start_probs = tf.gather_nd(start_probs, tf.stack([answer2support, start], 1))

    # only start probs are normalized on multi-paragraph, end probs conditioned on start only on per support level
    num_answers = tf.shape(answer_span)[0]
    is_aligned = tf.equal(tf.shape(end_scores)[0], num_answers)
    end_probs = tf.cond(
        is_aligned,
        lambda: tf.gather_nd(tf.nn.softmax(end_scores), tf.stack([tf.range(num_answers, dtype=tf.int32), end], 1)),
        lambda: tf.gather_nd(segment_softmax(end_scores, support2question), tf.stack([answer2support, end], 1))
    )

    answer2question = tf.gather(support2question, answer2support)
    # compute losses individually
    if use_sum:
        span_probs = tf.unsorted_segment_sum(
            start_probs, answer2question, num_questions) * tf.unsorted_segment_sum(
            end_probs, answer2question, num_questions)
    else:
        span_probs = tf.unsorted_segment_max(
            start_probs, answer2question, num_questions) * tf.unsorted_segment_max(
            end_probs, answer2question, num_questions)

    return -tf.reduce_mean(tf.log(tf.maximum(1e-6, span_probs + 1e-6)))



# -*- coding: utf-8 -*-

import logging
import os
import sys
from abc import abstractmethod
from functools import reduce
from typing import Iterable, Tuple, List, Mapping, Sequence

import numpy as np
import tensorflow as tf

from jack.core import JTReader, QASetting, Answer, Ports, ModelModule, SharedResources, TensorPort
from jack.core.reader import logger


class TFModelModule(ModelModule):
    """This class represents an  abstract ModelModule for tensroflow models which requires the implementation of
    a small set of methods that produce the TF graphs to create predictions and the training outputs,
    and define the ports.
    """

    def __init__(self, shared_resources: SharedResources, session=None):
        self.shared_resources = shared_resources
        if session is None:
            session_config = tf.ConfigProto(allow_soft_placement=True)
            session_config.gpu_options.allow_growth = True
            session = tf.Session(config=session_config)
        self.tf_session = session
        # will be set in setup
        self._tensors = None
        self._placeholders = None

    def __call__(self, batch: Mapping[TensorPort, np.ndarray],
                 goal_ports: List[TensorPort] = None) -> Mapping[TensorPort, np.ndarray]:
        """Runs a batch and returns values/outputs for specified goal ports.
        Args:
            batch: mapping from ports to values
            goal_ports: optional output ports, defaults to output_ports of this module will be returned

        Returns:
            A mapping from goal ports to tensors.
        """
        goal_ports = goal_ports or self.output_ports
        feed_dict = self.convert_to_feed_dict(batch)
        goal_tensors = {p: self.tensors[p] for p in goal_ports
                        if p in self.output_ports or p in self.training_output_ports}
        outputs = self.tf_session.run(goal_tensors, feed_dict)

        for p in goal_ports:
            if p not in outputs and p in batch:
                outputs[p] = batch[p]

        return outputs

    @abstractmethod
    def create_output(self, shared_resources: SharedResources,
                      input_tensors: Mapping[TensorPort, tf.Tensor]) -> \
            Mapping[TensorPort, tf.Tensor]:
        """
        This function needs to be implemented in order to define how the module produces
        output from input tensors corresponding to `input_ports`.

        Args:
            shared_resources: contains resources shared by modules, such as hyper-parameters or vocabularies.
            *input_tensors: a list of input tensors.

        Returns:
            mapping from defined output ports to their tensors.
        """
        raise NotImplementedError

    @abstractmethod
    def create_training_output(self, shared_resources: SharedResources,
                               training_input_tensors: Mapping[TensorPort, tf.Tensor]) \
            -> Mapping[TensorPort, tf.Tensor]:
        """
        This function needs to be implemented in order to define how the module produces tensors only used
        during training given tensors corresponding to the ones defined by `training_input_ports`, which might include
        tensors corresponding to ports defined by `output_ports`. This sub-graph should only be created during training.

        Args:
            shared_resources: contains resources shared by modules, such as hyper-parameters or vocabularies.
            training_input_tensors: a mapping from training input tensorports to tensors.

        Returns:
            mapping from defined training output ports to their tensors.
        """
        raise NotImplementedError

    def setup(self, is_training=True, reuse=False):
        """Sets up the module.

        This usually involves creating the actual tensorflow graph. It is expected to be called after the input module
        is set up and shared resources, such as the vocab, config, etc., are prepared already at this point.
        """
        old_train_variables = tf.trainable_variables()
        old_variables = tf.global_variables()

        with tf.variable_scope(self.shared_resources.config.get("name", "jtreader"),
                               initializer=tf.contrib.layers.xavier_initializer(), reuse=reuse):
            self._tensors = {p: p.create_tf_placeholder() for p in self.input_ports}
            output_tensors = self.create_output(
                self.shared_resources, {port: self._tensors[port] for port in self.input_ports})

            self._placeholders = dict(self._tensors)
            self._tensors.update(output_tensors)

            if is_training:
                self._placeholders.update((p, p.create_tf_placeholder()) for p in self.training_input_ports
                                          if p not in self._placeholders and p not in self._tensors)
                self._tensors.update(self._placeholders)
                input_target_tensors = {p: self._tensors.get(p, None) for p in self.training_input_ports}
                training_output_tensors = self.create_training_output(
                    self.shared_resources, {port: input_target_tensors[port] for port in self.training_input_ports})
                self._tensors.update(training_output_tensors)
        self._training_variables = [v for v in tf.trainable_variables() if v not in old_train_variables]
        self._saver = tf.train.Saver(self._training_variables, max_to_keep=1)
        self._variables = [v for v in tf.global_variables() if v not in old_variables]
        self.tf_session.run([v.initializer for v in self.variables])

        for var in tf.global_variables():
            is_trainable = var in tf.trainable_variables()
            logger.debug('Variable: {} (Trainable: {})'.format(var, is_trainable))

        # Sometimes we want to initialize (partially) with a pre-trained model
        load_dir = self.shared_resources.config.get('load_dir')
        if is_training and load_dir is not None:
            if not load_dir.endswith('model_module'):
                # path to a reader was provided
                load_dir = os.path.join(load_dir, 'model_module')
            # get all variables in the checkpoint file
            from tensorflow.python import pywrap_tensorflow
            reader = pywrap_tensorflow.NewCheckpointReader(load_dir)
            init_vars = []
            for n in reader.get_variable_to_shape_map().keys():
                found = False
                for v in self.variables:
                    if v.op.name == n:
                        found = True
                        init_vars.append(v)
                        break
                if not found:
                    logger.warn("Could not find variable %s in computation graph to restore from pretrained model." % n)

            saver = tf.train.Saver(init_vars)
            saver.restore(self.tf_session, load_dir)

    @property
    def placeholders(self) -> Mapping[TensorPort, tf.Tensor]:
        if hasattr(self, "_placeholders"):
            return self._placeholders
        else:
            logger.warn("Asking for placeholders without having setup this module. Returning None.")
            return None

    @property
    def tensors(self) -> Mapping[TensorPort, tf.Tensor]:
        if hasattr(self, "_tensors"):
            return self._tensors
        else:
            logger.warn("Asking for tensors without having setup this module. Returning None.")
            return None

    def store(self, path):
        self._saver.save(self.tf_session, path)

    def load(self, path):
        self._saver.restore(self.tf_session, path)

    @property
    def train_variables(self) -> Sequence[tf.Tensor]:
        return self._training_variables

    @property
    def variables(self) -> Sequence[tf.Tensor]:
        return self._variables

    def convert_to_feed_dict(self, mapping: Mapping[TensorPort, np.ndarray]) -> Mapping[tf.Tensor, np.ndarray]:
        result = {ph: mapping[port] for port, ph in self.placeholders.items() if port in mapping}
        return result


class TFReader(JTReader):
    """Tensorflow implementation of JTReader.

    A tensorflow reader reads inputs consisting of questions, supports and possibly candidates, and produces answers.
    It consists of three layers: input to tensor (input_module), tensor to tensor (model_module), and tensor to answer
    (output_model). These layers are called in-turn on a given input (list).
    """

    @property
    def model_module(self) -> TFModelModule:
        return super().model_module

    @property
    def session(self) -> tf.Session:
        """Returns: input module"""
        return self.model_module.tf_session

    def train(self, optimizer,
              training_set: Iterable[Tuple[QASetting, List[Answer]]],
              batch_size: int, max_epochs=10, hooks=tuple(),
              l2=0.0, clip=None, clip_op=tf.clip_by_value, summary_writer=None, **kwargs):
        """
        This method trains the reader (and changes its state).

        Args:
            optimizer: TF optimizer
            training_set: the training instances.
            batch_size: size of training batches
            max_epochs: maximum number of epochs
            hooks: TrainingHook implementations that are called after epochs and batches
            l2: whether to use l2 regularization
            clip: whether to apply gradient clipping and at which value
            clip_op: operation to perform for clipping
            summary_writer: summary writer
        """
        batches, loss, min_op, summaries = self._setup_training(
            batch_size, clip, optimizer, training_set, summary_writer, l2, clip_op, **kwargs)

        self._train_loop(min_op, loss, batches, hooks, max_epochs, summaries, summary_writer, **kwargs)

    def _setup_training(self, batch_size, clip, optimizer, training_set, summary_writer, l2, clip_op, **kwargs):
        global_step = tf.train.get_global_step()
        if global_step is None:
            global_step = tf.train.create_global_step()
        if not self._is_setup:
            # First setup shared resources, e.g., vocabulary. This depends on the input module.
            logger.info("Setting up model...")
            self.setup_from_data(training_set, is_training=True)
        logger.info("Preparing training data...")
        batches = self.input_module.batch_generator(training_set, batch_size, is_eval=False)
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
        loss = self.model_module.tensors[Ports.loss]
        summaries = None
        if summary_writer is not None:
            summaries = tf.summary.merge_all()
        if l2:
            loss += tf.add_n([tf.nn.l2_loss(v) for v in self.model_module.train_variables]) * l2
        if clip:
            gradients = optimizer.compute_gradients(loss)
            if clip_op == tf.clip_by_value:
                gradients = [(tf.clip_by_value(grad, clip[0], clip[1]), var)
                             for grad, var in gradients if grad is not None]
            elif clip_op == tf.clip_by_norm:
                gradients = [(tf.clip_by_norm(grad, clip), var)
                             for grad, var in gradients if grad is not None]
            min_op = optimizer.apply_gradients(gradients, global_step)
        else:
            min_op = optimizer.minimize(loss, global_step)

        variable_size = lambda v: reduce(lambda x, y: x * y, v.get_shape().as_list()) if v.get_shape() else 1
        num_params = sum(variable_size(v) for v in self.model_module.train_variables)
        logger.info("Number of parameters: %d" % num_params)

        # initialize non model variables like learning rate, optimizer vars ...
        self.session.run([v.initializer for v in tf.global_variables() if v not in self.model_module.variables])
        return batches, loss, min_op, summaries

    def _train_loop(self, optimization_op, loss_op, batches, hooks, max_epochs, summaries, summary_writer, **kwargs):
        logger.info("Start training...")
        for i in range(1, max_epochs + 1):
            for j, batch in enumerate(batches):
                feed_dict = self.model_module.convert_to_feed_dict(batch)

                if summaries is not None:
                    step, sums, current_loss, _ = self.session.run(
                        [tf.train.get_global_step(), summaries, loss_op, optimization_op], feed_dict=feed_dict)
                    summary_writer.add_summary(sums, step)
                else:
                    current_loss, _ = self.session.run([loss_op, optimization_op], feed_dict=feed_dict)

                for hook in hooks:
                    hook.at_iteration_end(i, current_loss, set_name='train')

            # calling post-epoch hooks
            for hook in hooks:
                hook.at_epoch_end(i)



# -*- coding: utf-8 -*-

from jack.core import *
from jack.core.data_structures import *
from jack.core.tensorflow import TFModelModule
from jack.readers.link_prediction import scores
from jack.util.map import numpify


class KnowledgeGraphEmbeddingInputModule(OnlineInputModule[List[List[int]]]):
    def __init__(self, shared_resources):
        self._kbp_rng = np.random.RandomState(0)
        super(KnowledgeGraphEmbeddingInputModule, self).__init__(shared_resources)

    def setup_from_data(self, data: Iterable[Tuple[QASetting, List[Answer]]]):
        triples = [tuple(x[0].question.split()) for x in data]

        entity_set = {s for [s, _, _] in triples} | {o for [_, _, o] in triples}
        predicate_set = {p for [_, p, _] in triples}

        entity_to_index = {entity: index for index, entity in enumerate(entity_set, start=1)}
        predicate_to_index = {predicate: index for index, predicate in enumerate(predicate_set, start=1)}

        self.shared_resources.entity_to_index = entity_to_index
        self.shared_resources.predicate_to_index = predicate_to_index

        self.shared_resources.nb_entities = max(self.shared_resources.entity_to_index.values()) + 1
        self.shared_resources.nb_predicates = max(self.shared_resources.predicate_to_index.values()) + 1


    def preprocess(self, questions: List[QASetting], answers: Optional[List[List[Answer]]] = None,
                   is_eval: bool = False) -> List[List[int]]:
        """Converts questions to triples."""
        triples = []
        for qa_setting in questions:
            s, p, o = qa_setting.question.split()
            s_idx = self.shared_resources.entity_to_index.get(s, 0)
            o_idx = self.shared_resources.entity_to_index.get(o, 0)
            p_idx = self.shared_resources.predicate_to_index.get(p, 0)
            triples.append([s_idx, p_idx, o_idx])

        return triples

    def create_batch(self, triples: List[List[int]],
                     is_eval: bool, with_answers: bool) -> Mapping[TensorPort, np.ndarray]:
        _triples = list(triples)

        if with_answers:
            target = [1] * len(_triples)

        nb_entities = self.shared_resources.nb_entities
        nb_predicates = self.shared_resources.nb_predicates

        if with_answers:
            for i in range(len(_triples)):
                s, p, o = triples[i]

                for _ in range(self.shared_resources.config.get('num_negative', 1)):

                    random_subject_index = self._kbp_rng.randint(0, nb_entities)
                    random_object_index = self._kbp_rng.randint(0, nb_predicates)

                    _triples.append([random_subject_index, p, o])
                    _triples.append([s, p, random_object_index])

                    target.append(0)
                    target.append(0)

        xy_dict = {Ports.Input.question: _triples}

        if with_answers:
            xy_dict[Ports.Target.target_index] = target

        return numpify(xy_dict)

    @property
    def output_ports(self) -> List[TensorPort]:
        return [Ports.Input.question]

    @property
    def training_ports(self) -> List[TensorPort]:
        return [Ports.Target.target_index]


class KnowledgeGraphEmbeddingModelModule(TFModelModule):
    def __init__(self, *args, model_name='DistMult', **kwargs):
        super().__init__(*args, **kwargs)
        self.model_name = model_name

    @property
    def input_ports(self) -> List[TensorPort]:
        return [Ports.Input.question]

    @property
    def output_ports(self) -> List[TensorPort]:
        return [Ports.Prediction.logits]

    @property
    def training_input_ports(self) -> List[TensorPort]:
        return [Ports.Target.target_index, Ports.Prediction.logits]

    @property
    def training_output_ports(self) -> List[TensorPort]:
        return [Ports.loss]

    def create_training_output(self, shared_resources: SharedResources, input_tensors) \
            -> Mapping[TensorPort, tf.Tensor]:
        tensors = TensorPortTensors(input_tensors)
        losses = tf.nn.sigmoid_cross_entropy_with_logits(logits=tensors.logits,
                                                         labels=tf.to_float(tensors.target_index))
        loss = tf.reduce_mean(losses, axis=0)
        return {Ports.loss: loss}

    def create_output(self, shared_resources: SharedResources, input_tensors) -> Mapping[TensorPort, tf.Tensor]:
        tensors = TensorPortTensors(input_tensors)
        with tf.variable_scope('knowledge_graph_embedding'):
            embedding_size = shared_resources.config['repr_dim']

            nb_entities = max(shared_resources.entity_to_index.values()) + 1
            nb_predicates = max(shared_resources.predicate_to_index.values()) + 1

            entity_embeddings = tf.get_variable('entity_embeddings',
                                                shape=[nb_entities, embedding_size],
                                                initializer=tf.contrib.layers.xavier_initializer(),
                                                dtype='float32')

            predicate_embeddings = tf.get_variable('predicate_embeddings',
                                                   shape=[nb_predicates, embedding_size],
                                                   initializer=tf.contrib.layers.xavier_initializer(),
                                                   dtype='float32')

            subject_idx = tensors.question[:, 0]
            predicate_idx = tensors.question[:, 1]
            object_idx = tensors.question[:, 2]

            subject_emb = tf.nn.embedding_lookup(entity_embeddings, subject_idx, max_norm=1.0)
            predicate_emb = tf.nn.embedding_lookup(predicate_embeddings, predicate_idx)
            object_emb = tf.nn.embedding_lookup(entity_embeddings, object_idx, max_norm=1.0)

            assert self.model_name is not None

            model_class = scores.get_function(self.model_name)
            model = model_class(
                subject_embeddings=subject_emb,
                predicate_embeddings=predicate_emb,
                object_embeddings=object_emb)

            logits = model()

        return {
            Ports.Prediction.logits: logits
        }


class KnowledgeGraphEmbeddingOutputModule(OutputModule):
    def setup(self):
        pass

    @property
    def input_ports(self) -> List[TensorPort]:
        return [Ports.Prediction.logits]

    def __call__(self, questions: Sequence[QASetting], tensors: Mapping[TensorPort, np.array]) \
            -> Sequence[Sequence[Answer]]:
        # len(inputs) == batch size
        # logits: [batch_size, max_num_candidates]
        logits = tensors[Ports.Prediction.logits]
        results = []
        for index_in_batch, question in enumerate(questions):
            score = logits[index_in_batch]
            results.append([Answer(question.question, score=score)])
        return results



"""Implementation of loaders for common datasets."""

import json

from jack.core.data_structures import *
from jack.io.SNLI2jtr import convert_snli
from jack.io.SQuAD2jtr import convert_squad

loaders = dict()


def _register(name):
    def _decorator(f):
        loaders[name] = f
        return f

    return _decorator


@_register('jack')
def load_jack(path, max_count=None):
    """
    This function loads a jack json file from a specific location.
    Args:
        path: the location to load from.
        max_count: how many instances to load at most

    Returns:
        A list of input-answer pairs.

    """
    # We load json directly instead
    with open(path) as f:
        jtr_data = json.load(f)

    return jack_to_qasetting(jtr_data, max_count)


@_register('squad')
def load_squad(path, max_count=None):
    """
    This function loads a squad json file from a specific location.
    Args:
        path: the location to load from.
        max_count: how many instances to load at most

    Returns:
        A list of input-answer pairs.
    """
    # We load to jtr dict and convert to qa settings for now
    jtr_data = convert_squad(path)
    return jack_to_qasetting(jtr_data, max_count)


@_register('snli')
def load_snli(path, max_count=None):
    """
    This function loads a jack json file with labelled answers from a specific location.
    Args:
        path: the location to load from.
        max_count: how many instances to load at most

    Returns:
        A list of input-answer pairs.
    """
    # We load to jtr dict and convert to qa settings for now
    jtr_data = convert_snli(path)
    return jack_to_qasetting(jtr_data, max_count)



# -*- coding: utf-8 -*-

from jack.io.embeddings.embeddings import Embeddings, load_embeddings
from jack.io.embeddings.glove import load_glove

__all__ = [
    'Embeddings',
    'load_embeddings'
    'load_word2vec',
    'get_word2vec_vocabulary',
    'load_glove',
]



import tensorflow as tf


def prediction_layer(hypothesis, hypothesis_length, premise, premise_length, num_classes, is_eval,
                     repr_dim=100, activation=None, module='max_avg_mlp', dropout=0.0, **kwargs):
    if module == 'max_avg_mlp':
        return max_avg_pool_prediction_layer(
            repr_dim, num_classes, hypothesis, hypothesis_length, premise, premise_length, is_eval, activation, dropout)
    elif module == 'max_mlp':
        return max_pool_prediction_layer(
            repr_dim, num_classes, hypothesis, hypothesis_length, premise, premise_length, is_eval, activation, dropout)
    elif module == 'max_mlp_hypothesis':
        return max_pool_prediction_layer_hypothesis(
            repr_dim, num_classes, hypothesis, hypothesis_length, is_eval, activation, dropout)
    elif module == 'max_interaction_mlp':
        return max_pool_interaction_prediction_layer(
            repr_dim, num_classes, hypothesis, hypothesis_length, premise, premise_length, is_eval, activation, dropout)
    else:
        raise ValueError("Unknown answer layer type: %s" % module)


def _mask(hypothesis, hypothesis_length, premise, premise_length):
    p_mask = tf.sequence_mask(premise_length, tf.shape(premise)[1], dtype=tf.float32)
    h_mask = tf.sequence_mask(hypothesis_length, tf.shape(hypothesis)[1], dtype=tf.float32)

    premise *= tf.expand_dims(p_mask, 2)
    hypothesis *= tf.expand_dims(h_mask, 2)
    return hypothesis, premise


def max_avg_pool_prediction_layer(size, num_classes, hypothesis, hypothesis_length, premise, premise_length, is_eval,
                                  activation=tf.tanh, dropout=0.0):
    hypothesis, premise = _mask(hypothesis, hypothesis_length, premise, premise_length)

    p_max = tf.reduce_max(premise, axis=1)
    p_avg = tf.reduce_sum(premise, axis=1) / tf.expand_dims(tf.to_float(premise_length), 1)

    h_max = tf.reduce_max(hypothesis, axis=1)
    h_avg = tf.reduce_sum(hypothesis, axis=1) / tf.expand_dims(tf.to_float(hypothesis_length), 1)

    inputs = tf.concat([p_max, p_avg, h_max, h_avg], 1)
    if dropout:
        inputs = tf.cond(is_eval, lambda: inputs, lambda: tf.nn.dropout(inputs, 1.0 - dropout))
    hidden = tf.layers.dense(inputs, size, activation)
    if dropout:
        hidden = tf.cond(is_eval, lambda: hidden, lambda: tf.nn.dropout(hidden, 1.0 - dropout))
    logits = tf.layers.dense(hidden, num_classes)

    return logits


def max_pool_prediction_layer(size, num_classes, hypothesis, hypothesis_length, premise, premise_length, is_eval,
                              activation=tf.tanh, dropout=0.0):
    hypothesis, premise = _mask(hypothesis, hypothesis_length, premise, premise_length)

    p_max = tf.reduce_max(premise, axis=1)

    h_max = tf.reduce_max(hypothesis, axis=1)

    inputs = tf.concat([p_max, h_max], 1)
    if dropout:
        inputs = tf.cond(is_eval, lambda: inputs, lambda: tf.nn.dropout(inputs, 1.0 - dropout))
    hidden = tf.layers.dense(inputs, size, activation)
    if dropout:
        hidden = tf.cond(is_eval, lambda: hidden, lambda: tf.nn.dropout(hidden, 1.0 - dropout))
    logits = tf.layers.dense(hidden, num_classes)

    return logits


def max_pool_interaction_prediction_layer(size, num_classes, hypothesis, hypothesis_length, premise, premise_length,
                                          is_eval, activation=tf.tanh, dropout=0.0):
    hypothesis, premise = _mask(hypothesis, hypothesis_length, premise, premise_length)

    p_max = tf.reduce_max(premise, axis=1)
    h_max = tf.reduce_max(hypothesis, axis=1)

    inputs = tf.concat([p_max, h_max, p_max - h_max, p_max * h_max], 1)
    if dropout:
        inputs = tf.cond(is_eval, lambda: inputs, lambda: tf.nn.dropout(inputs, 1.0 - dropout))
    hidden = tf.layers.dense(inputs, size, activation)
    if dropout:
        hidden = tf.cond(is_eval, lambda: hidden, lambda: tf.nn.dropout(hidden, 1.0 - dropout))
    logits = tf.layers.dense(hidden, num_classes)

    return logits


def max_pool_prediction_layer_hypothesis(size, num_classes, hypothesis, hypothesis_length, is_eval,
                                         activation=tf.tanh, dropout=0.0):
    h_mask = tf.sequence_mask(hypothesis_length, tf.shape(hypothesis)[1], dtype=tf.float32)
    hypothesis *= tf.expand_dims(h_mask, 2)

    h_max = tf.reduce_max(hypothesis, axis=1)

    hidden = tf.layers.dense(h_max, size, activation)
    if dropout:
        hidden = tf.cond(is_eval, lambda: hidden, lambda: tf.nn.dropout(hidden, 1.0 - dropout))
    logits = tf.layers.dense(hidden, num_classes)

    return logits



import json
import io
import random

def convert_scienceCloze_to_jtr(scienceQAFile):

    instances = []

    f = io.open(scienceQAFile, "r", encoding="utf-8")

    for l in f:
        l = l.strip().lower().split("\t")  # do the lower case preprocessing here
        try:
            quest, answs, cands, context, contextID = l
        except ValueError:
            print(l)
            continue

        context = context[2:-2].split('\', \'')

        support = []
        for i, c in enumerate(context):
            support.append({"id": contextID + "_" + str(i), "text": c})
        candidates = cands[2:-2].split('\', \'')

        qdict = {
            'question': quest,
            'candidates': [
                {
                    'text': cand
                } for cand in candidates
                ],
            'answers': [{'text': answs}]
        }
        qset_dict = {
            'support': support,
            'questions': [qdict]
        }

        instances.append(qset_dict)


    instances.append(qset_dict)
    random.shuffle(instances)

    corpus_dict = {
        'meta': "scienceQA.json",
        'instances': instances
    }

    f.close()

    return corpus_dict



if __name__ == "__main__":
    corpus = convert_scienceCloze_to_jtr("../data/scienceQA/clozeSummaryLocal_test.txt")
    with open("../data/scienceQA/scienceQA_clozeSummaryLocal_test.json", 'w') as outfile:
        json.dump(corpus, outfile, indent=2, ensure_ascii=False)


# -*- coding: utf-8 -*-

import tensorflow as tf


def mask_3d(sequences, sequence_lengths, mask_value, dimension=2):
    """
    Given a batch of matrices, each with shape m x n, mask the values in each
    row after the positions indicated in sentence_sizes.
    This function is supposed to mask the last columns in the raw attention
    matrix (e_{i, j}) in cases where the sentence2 is smaller than the
    maximum.
    
    Args:
        sequences: tensor with shape (batch_size, m, n)
        sequence_lengths: tensor with shape (batch_size) containing the sentence sizes that
           should be limited
        mask_value: scalar value to assign to items after sentence size
        dimension: over which dimension to mask values
    Returns:
        A tensor with the same shape as `values`
    """
    if dimension == 1:
        sequences = tf.transpose(sequences, [0, 2, 1])
    time_steps1, time_steps2 = tf.shape(sequences)[1], tf.shape(sequences)[2]
    ones = tf.ones_like(sequences, dtype=tf.int32)
    pad_values = mask_value * tf.cast(ones, tf.float32)
    mask = tf.sequence_mask(sequence_lengths, time_steps2)
    # mask is (batch_size, sentence2_size). we have to tile it for 3d
    mask3d = tf.tile(tf.expand_dims(mask, 1), (1, time_steps1, 1))
    masked = tf.where(mask3d, sequences, pad_values)
    return tf.transpose(masked, [0, 2, 1]) if dimension == 1 else masked



import tensorflow as tf

from jack.util.tf import attention, sequence_encoder
from jack.util.tf.misc import mask_for_lengths


def interaction_layer(seq1, seq1_length, seq2, seq2_length, seq1_to_seq2, seq2_to_seq1,
                      module='attention_matching', name='interaction_layer', reuse=False, concat=True,
                      repr_dim=None, **kwargs):
    with tf.variable_scope(name, reuse=reuse):
        if seq1_to_seq2 is not None:
            seq2 = tf.gather(seq2, seq1_to_seq2)
            seq2_length = tf.gather(seq2_length, seq1_to_seq2)
        if module == 'attention_matching':
            out = attention_matching_layer(seq1, seq1_length, seq2, seq2_length, seq2_to_seq1, **kwargs)
        elif module == 'bidaf':
            out = bidaf_layer(seq1, seq1_length, seq2, seq2_length, **kwargs)
        elif module == 'coattention':
            if kwargs.get('encoder') and 'repr_dim' not in kwargs['encoder']:
                kwargs['encoder']['repr_dim'] = repr_dim
            out = coattention_layer(
                seq1, seq1_length, seq2, seq2_length, **kwargs)
        else:
            raise ValueError("Unknown interaction type: %s" % module)

    if concat:
        out = tf.concat([seq1, out], 2)
    return out


def bidaf_layer(seq1, seq1_length, seq2, seq2_length, seq2_to_seq1=None, **kwargs):
    """Encodes seq1 conditioned on seq2, e.g., using word-by-word attention."""
    attn_scores, attn_probs, seq2_weighted = attention.diagonal_bilinear_attention(
        seq1, seq2, seq2_length, False, seq2_to_seq1=seq2_to_seq1)

    attn_scores += tf.expand_dims(mask_for_lengths(seq1_length, tf.shape(attn_scores)[1]), 2)

    max_seq1 = tf.reduce_max(attn_scores, 2)
    seq1_attention = tf.nn.softmax(max_seq1, 1)
    seq1_weighted = tf.einsum('ij,ijk->ik', seq1_attention, seq1)
    seq1_weighted = tf.expand_dims(seq1_weighted, 1)
    seq1_weighted = tf.tile(seq1_weighted, [1, tf.shape(seq1)[1], 1])

    return tf.concat([seq2_weighted, seq1 * seq2_weighted, seq1 * seq1_weighted], 2)


def attention_matching_layer(seq1, seq1_length, seq2, seq2_length, seq2_to_seq1=None,
                             attn_type='diagonal_bilinear', key_value_attn=False,
                             scaled=True, with_sentinel=False, repr_dim=None, **kwargs):
    """Encodes seq1 conditioned on seq2, e.g., using word-by-word attention."""
    if attn_type == 'bilinear':
        _, _, attn_states = attention.bilinear_attention(seq1, seq2, seq2_length, scaled, with_sentinel,
                                                         seq2_to_seq1=seq2_to_seq1)
    elif attn_type == 'dot':
        _, _, attn_states = attention.dot_attention(seq1, seq2, seq2_length, scaled, with_sentinel,
                                                    seq2_to_seq1=seq2_to_seq1)
    elif attn_type == 'diagonal_bilinear':
        _, _, attn_states = attention.diagonal_bilinear_attention(seq1, seq2, seq2_length, scaled,
                                                                  with_sentinel, seq2_to_seq1=seq2_to_seq1)
    elif attn_type == 'mlp':
        _, _, attn_states = attention.mlp_attention(seq1.get_shape()[-1].value, tf.nn.relu, seq1, seq2,
                                                    seq2_length, with_sentinel, seq2_to_seq1=seq2_to_seq1)
    else:
        raise ValueError("Unknown attention type: %s" % attn_type)

    return attn_states


def coattention_layer(seq1, seq1_length, seq2, seq2_length,
                      attn_type='diagonal_bilinear', scaled=True, with_sentinel=False, seq2_to_seq1=None,
                      num_layers=1, encoder=None, **kwargs):
    """Encodes seq1 conditioned on seq2, e.g., using word-by-word attention."""
    if attn_type == 'bilinear':
        attn_fun = attention.bilinear_attention
    elif attn_type == 'dot':
        attn_fun = attention.dot_attention
    elif attn_type == 'diagonal_bilinear':
        attn_fun = attention.diagonal_bilinear_attention
    else:
        raise ValueError("Unknown attention type: %s" % attn_type)

    _, _, attn_states1, attn_states2, co_attn_state = attention.coattention(
        seq1, seq1_length, seq2, seq2_length, scaled, with_sentinel, attn_fun)

    if num_layers < 2:
        out = tf.concat([attn_states1, co_attn_state], 2)
    else:
        seq1, attn_states1, attn_states2, co_attn_state = [], [attn_states1], [attn_states2], [co_attn_state]
        for i in range(1, num_layers):
            with tf.variable_scope(str(i)):
                enc_1 = sequence_encoder.encoder(
                    attn_states1[-1], seq1_length, name='encoder1', **encoder)
                enc_2 = sequence_encoder.encoder(
                    attn_states2[-1], seq2_length, name='encoder2', **encoder)
                seq1.append(enc_1)
                _, _, new_attn_states1, new_attn_states2, new_co_attn_state = attention.coattention(
                    enc_1, seq1_length, enc_2, seq2_length, scaled, with_sentinel, attn_fun, seq2_to_seq1=seq2_to_seq1)
                attn_states1.append(new_attn_states1)
                attn_states2.append(new_attn_states2)
                co_attn_state.append(new_co_attn_state)
        out = tf.concat(seq1 + attn_states1 + co_attn_state, 2)

    return out



import logging
import re
import string
from collections import Counter

logger = logging.getLogger(__name__)


def evaluate(reader, dataset, batch_size):
    answers = reader.process_dataset(dataset, batch_size, silent=False)

    f1 = exact_match = 0
    for pa, (q, ass) in zip(answers, dataset):
        ground_truth = [a.text for a in ass]
        f1 += metric_max_over_ground_truths(f1_score, pa.text, ground_truth)
        exact_match += metric_max_over_ground_truths(exact_match_score, pa.text, ground_truth)

    f1 /= len(answers)
    exact_match /= len(answers)

    return {'F1': f1, 'Exact': exact_match}


def normalize_answer(s):
    """Lower text and remove punctuation, articles and extra whitespace."""

    def remove_articles(text):
        return re.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1_score(prediction, ground_truth):
    prediction_tokens = normalize_answer(prediction).split()
    ground_truth_tokens = normalize_answer(ground_truth).split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def exact_match_score(prediction, ground_truth):
    return normalize_answer(prediction) == normalize_answer(ground_truth)


def metric_max_over_ground_truths(metric_fn, prediction, ground_truths):
    scores_for_ground_truths = [0.0]
    for ground_truth in ground_truths:
        score = metric_fn(prediction, ground_truth)
        scores_for_ground_truths.append(score)
    return max(scores_for_ground_truths)



# -*- coding: utf-8 -*-

from jack.eval import extractive_qa, link_prediction, classification

evaluators = {
    'extractive_qa': extractive_qa.evaluate,
    'link_prediction': link_prediction.evaluate,
    'classification': None
}


def evaluate_reader(reader, dataset, batch_size):
    from jack.readers.implementations import extractive_qa_readers, classification_readers, link_prediction_readers
    reader_name = reader.shared_resources.config.get('reader')
    if reader_name in extractive_qa_readers:
        return extractive_qa.evaluate(reader, dataset, batch_size)
    elif reader_name in link_prediction_readers:
        return link_prediction.evaluate(reader, dataset, batch_size)
    elif reader_name in classification_readers:
        return classification.evaluate(reader, dataset, batch_size)


def pretty_print_results(d, prefix=''):
    for k, v in sorted(d.items(), key=lambda x: x[0]):
        if isinstance(v, dict):
            print(prefix + k + ":")
            pretty_print_results(v, prefix + '\t')
        elif '\n' in str(v):
            print(prefix + k + ":")
            print(str(v).replace('\n', '\n' + prefix + '\t'))
        else:
            print(prefix + k + ":", str(v))



from typing import Sequence

from jack.core import Ports, TensorPort, TensorPortTensors
from jack.core.tensorflow import TFModelModule
from jack.readers.extractive_qa.shared import XQAPorts
from jack.util.tf.xqa import xqa_crossentropy_loss


class AbstractXQAModelModule(TFModelModule):
    _input_ports = [XQAPorts.emb_question, XQAPorts.question_length,
                    XQAPorts.emb_support, XQAPorts.support_length, XQAPorts.support2question,
                    # char embedding inputs
                    XQAPorts.word_chars, XQAPorts.word_char_length,
                    XQAPorts.question_batch_words, XQAPorts.support_batch_words,
                    # feature input
                    XQAPorts.word_in_question,
                    # optional input, provided only during training
                    XQAPorts.correct_start, XQAPorts.answer2support_training,
                    XQAPorts.is_eval]

    _output_ports = [XQAPorts.start_scores, XQAPorts.end_scores,
                     XQAPorts.answer_span]
    _training_input_ports = [XQAPorts.start_scores, XQAPorts.end_scores,
                             XQAPorts.answer_span_target, XQAPorts.answer2support_training, XQAPorts.support2question]
    _training_output_ports = [Ports.loss]

    @property
    def output_ports(self) -> Sequence[TensorPort]:
        return self._output_ports

    @property
    def input_ports(self) -> Sequence[TensorPort]:
        return self._input_ports

    @property
    def training_input_ports(self) -> Sequence[TensorPort]:
        return self._training_input_ports

    @property
    def training_output_ports(self) -> Sequence[TensorPort]:
        return self._training_output_ports

    def create_training_output(self, shared_resources, input_tensors):
        tensors = TensorPortTensors(input_tensors)
        return {
            Ports.loss: xqa_crossentropy_loss(tensors.start_scores, tensors.end_scores,
                                              tensors.answer_span_target, tensors.answer2support,
                                              tensors.support2question,
                                              use_sum=shared_resources.config.get('loss', 'sum') == 'sum')
        }



# -*- coding: utf-8 -*-

from abc import abstractmethod
from typing import Sequence, Mapping

import numpy as np

from jack.core.data_structures import QASetting, Answer
from jack.core.tensorport import TensorPort


class OutputModule:
    """
    An output module takes the output (numpy) tensors of the model module and turns them into
    jack data structures.
    """

    @property
    @abstractmethod
    def input_ports(self) -> Sequence[TensorPort]:
        """Returns: correspond to a subset of output ports of model module."""
        raise NotImplementedError

    @abstractmethod
    def __call__(self, questions: Sequence[QASetting], tensors: Mapping[TensorPort, np.array]) \
            -> Sequence[Answer]:
        """
        Process the tensors corresponding to the defined `input_ports` for a batch to produce a list of answers.
        The module has access to the original inputs.
        Args:
            questions:
            prediction:

        Returns:

        """
        raise NotImplementedError

    @abstractmethod
    def setup(self):
        pass

    def store(self, path):
        """Store the state of this module. Default is that there is no state, so nothing to store."""
        pass

    def load(self, path):
        """Load the state of this module. Default is that there is no state, so nothing to load."""
        pass



# -*- coding: utf-8 -*-

import sys

import tensorflow as tf


def negative_l1_distance(x1, x2, axis=1):
    """
    Negative L1 Distance.

    .. math:: L = - \\sum_i \\abs(x1_i - x2_i)

    Args:
        x1: First term.
        x2: Second term.
        axis: Reduction Indices.

    Returns:
        Similarity Value.
    """
    distance = tf.reduce_sum(tf.abs(x1 - x2), axis=axis)
    return - distance


def negative_l2_distance(x1, x2, axis=1):
    """
    Negative L2 Distance.

    .. math:: L = - \\sqrt{\\sum_i (x1_i - x2_i)^2}

    Args:
        x1: First term.
        x2: Second term.
        axis: Reduction Indices.

    Returns:
        Similarity Value.
    """

    distance = tf.sqrt(tf.reduce_sum(tf.square(x1 - x2), axis=axis))
    return - distance


def negative_square_l2_distance(x1, x2, axis=1):
    """
    Negative Square L2 Distance.

    .. math:: L = - \\sum_i (x1_i - x2_i)^2

    Args:
        x1: First term.
        x2: Second term.
        axis: Reduction Indices.

    Returns:
        Similarity Value.
    """
    distance = tf.reduce_sum(tf.square(x1 - x2), axis=axis)
    return - distance


def dot_product(x1, x2, axis=1):
    """
    Dot Product.

    .. math:: L = \\sum_i x1_i x2_i

    Args:
        x1: First term.
        x2: Second term.
        axis: Reduction Indices.

    Returns:
        Similarity Value.
    """

    similarity = tf.reduce_sum(x1 * x2, axis=axis)
    return similarity


# Aliases
l1 = L1 = negative_l1_distance
l2 = L2 = negative_l2_distance
l2_sqr = L2_SQR = negative_square_l2_distance
dot = DOT = dot_product


def get_function(function_name):
    this_module = sys.modules[__name__]
    if not hasattr(this_module, function_name):
        raise ValueError('Unknown similarity function: {}'.format(function_name))
    return getattr(this_module, function_name)



# -*- coding: utf-8 -*-

import gzip
import numpy as np

import logging

logger = logging.getLogger(__name__)


def load_word2vec(filename, vocab=None, normalise=True):
    """Loads a word2vec file and merges existing vocabulary.

    Args:
        filename (string): Path to the word2vec file.
        vocab (Vocabulary=None): Existing vocabulary to be merged.
        normalise (bool=True): If the word embeddings should be unit
                  normalized or not.
    Returns:
        return_vocab (dict), lookup (matrix): The dict is a word2idx dict and
        the lookup matrix is the matrix of embedding vectors.
    """
    logger.info("Loading word2vec vectors ..")
    with gzip.open(filename, 'rb') as f:
        vec_n, vec_size = map(int, f.readline().split())
        byte_size = vec_size * 4
        lookup = np.empty([vocab.get_size() if vocab is not None else vec_n, vec_size], dtype=np.float32)
        word2idx = {}
        idx = 0
        for n in range(vec_n):
            word = b''
            while True:
                c = f.read(1)
                if c == b' ':
                    break
                else:
                    word += c

            word = word.decode('utf-8')
            vector = np.fromstring(f.read(byte_size), dtype=np.float32)
            if vocab is None or vocab.contains_word(word):
                word2idx[word] = idx
                lookup[idx] = _normalise(vector) if normalise else vector
                idx += 1

    lookup.resize([idx, vec_size])
    logger.info('Loading word2vec vectors completed.')
    return word2idx, lookup


def _normalise(x):
    """Unit normalize x with L2 norm."""
    return (1.0 / np.linalg.norm(x, ord=2)) * x


def get_word2vec_vocabulary(fname):
    """Loads word2vec file and returns the vocabulary as dict word2idx."""
    voc, _ = load_word2vec(fname)
    return voc


if __name__ == "__main__":
    pickle_tokens = False
    vocab, _ = load_word2vec('../../data/word2vec/GoogleNews-vectors-negative300.bin.gz')

    # pickle token set
    if pickle_tokens:
        import pickle
        w2v_words = set(vocab.get_all_words())
        pickle.dump(w2v_words, open('./data/w2v_tokens.pickle', 'wb'))



# -*- coding: utf-8 -*-

import tensorflow as tf


def fully_connected_projection(inputs, output_size):
    """Projects inputs onto target dimension. Returns logits, loss, and argmax.

    Creates fully connected projection layer. Then applies cross entropy
    softmax to get the loss. Calculate predictions via argmax.
    Args:
        inputs (tensor): Input into the projection layer.
        output_size (int): Size of the targets (used in projection layer).
    """
    init = tf.contrib.layers.xavier_initializer(uniform=True) #uniform=False for truncated normal
    logits = tf.contrib.layers.fully_connected(inputs, output_size, weights_initializer=init, activation_fn=None)
    return logits



# -*- coding: utf-8 -*-

import random

from itertools import islice
from typing import TypeVar, List, Iterator, Optional

import numpy as np

from jack.util.map import numpify
from jack.util.random import DefaultRandomState

rs = DefaultRandomState(1337)


class GeneratorWithRestart(object):
    def __init__(self, iterator):
        self.iterator = iterator

    def __iter__(self):
        return self.iterator()


def get_buckets(data, order, structure):
    """
    Generates mapping between data instances and bucket-ID's.

    `data`: dict of nested sequences in which each top-level sequence has the same length,
        and all inner sequences have the __len__ attribute.
    `order`: (None or) tuple with data keys used for bucketing
        For example:
        ```list(data.keys()) = ["sentences1", "lengths1", "sentences2", "lengths2", "targets"]```
        and we want bucketing according to the lengths of inner sequences in "sentences1" and "sentences2":
        `order = ("sentences1", "sentences2")` performs bucketing on "sentences1", and within each bucket,
        again creates buckets according to "sentences2"
        (automatic bucketing will result in different "sentences2" bucket boundaries
        within each bucket according to "sentences1").
        `order = ("sentences2", "sentences1")`: vice versa, with "sentences2" for highest-level buckets
    `structure`: (None or) sequence with same length as `order`, each element is an integer or a list of integers
        For each position:
            - integer: denotes number of buckets, to be determined automatically
            - list: determines bucket boundaries. E.g.: [10, 20, 30] will result in 4 buckets
              (1) lengths 0-10, (2) lengths 11-20, (3) lengths 21-30, (4) lengths > 30
        For example:
        `order` = ("sentences1", "sentences2") and `structure` = (3, [10]) generates 6 buckets:
        within each of 3 partitions based on "sentences1",
        there is a bucket with instances of "sentences2" with length 10 or less,
        and one for lengths > 10.

    Returns:
        buckets2ids, ids2buckets
        dicts that map instance-id (index along 1st dimension of values in data) to bucket-id,
        and vice versa.
    """
    assert isinstance(data, dict)

    n_tot = len(list(data.values())[0])
    if order is None or structure is None:
        # all in 1 bucket, with id '(0)'
        buckets2ids = {'(0)': list(range(n_tot))}
        ids2buckets = dict(zip(list(range(n_tot)), ['(0)'] * n_tot))
        return buckets2ids, ids2buckets

    def _chunk(it, size):
        """returns iterator of chunks (tuples) from it (input iterator), with given size (last one may be shorter)"""
        it = iter(it)
        return iter(lambda: tuple(islice(it, size)), ())

    def _partition(_buckets2ids, _order, _structure):
        """update _buckets2ids according to _order and _structure"""
        # update all current buckets according to first item in _order and _structure
        buckets2ids_new = {}
        for bid, ids in sorted(_buckets2ids.items(), key=lambda x: x[0]):
            lengths = [len(data[_order[0]][id]) for id in ids]
            sorted_ids_lengths = sorted(zip(ids, lengths), key=lambda x: x[1])
            if isinstance(_structure[0], int):  # automatic bucketing
                size = len(lengths) // _structure[0] if len(lengths) % _structure[0] == 0 \
                    else 1 + (len(lengths) // _structure[0])
                buckets = list(_chunk([tup[0] for tup in sorted_ids_lengths], size))
            else:  # structure_is sequence of ints
                struct = list(sorted(_structure[0])) + [np.inf]
                bin_max, struct = struct[0], struct[1:]
                buckets = [[]]
                for id, l in sorted_ids_lengths:
                    if l > bin_max:  # never happens when bin_max = np.inf
                        bin_max, struct = struct[0], struct[1:]
                        buckets.append([])
                    buckets[-1].append(id)
            buckets2ids_new.update({tuple(list(bid) + [i]): list(bucket) for i, bucket in enumerate(buckets)})
        # call again if _order and _structure have more than 1 item
        if len(_order) > 1:
            buckets2ids_new = _partition(buckets2ids_new, _order[1:], _structure[1:])

        buckets2ids_new = {bid: bucket for bid, bucket in buckets2ids_new.items() if len(bucket) > 0}
        return buckets2ids_new

    buckets2ids = _partition({(): list(range(n_tot))}, order, structure)
    buckets2ids = {str(bid): buckets2ids[bid] for bid in buckets2ids}  # make bucket-ids strings (for random.choice)

    ids2buckets = {}
    for bid, bucket in buckets2ids.items():
        ids2buckets.update({id: bid for id in bucket})
    return buckets2ids, ids2buckets


def get_batches(data, batch_size=32, pad=0, bucket_order=None, bucket_structure=None, exact_epoch=False):
    """
    Creates generator that batches `data`.
    To avoid biases, it is advised to keep `bucket_order=None` and `bucket_structure=None` if computationally possible.
    (which will sample batches from all instances)

    Args:
        `data`: dict with (multi-dimensional) numpy arrays or (nested) lists;
            first inner dimension (`num_instances`) should be the same over all data values.
        `batch_size`: the desired batch size
        `pad`: padding symbol in case data contains lists of lists of different sizes
        `bucket_order`: argument `order` in get_buckets (list with keys); `None` if no bucketing
        `bucket_structure`: argument `structure` in get_buckets; `None` if no bucketing
        `exact_epoch`: if set to `True`, final batch per bucket may be smaller, but each instance will be seen exactly
            once during training. Default: `False`, to be certain during training
            that each instance per batch gets same weight in the total loss
            (but not all instances are observed per epoch if bucket sizes are no multiple of `batch_size`).

    Returns:
        a generator that generates a dict with same keys as `data`, and
        as values data batches consisting of `[batch_size x num_instances]` 2D numpy tensors
        (1st dimension is at most `batch_size` but may be smaller to cover all instances exactly once per epoch,
        if `exact_epoch=True`)
     """
    assert isinstance(data, dict)

    data0 = list(data.values())[0]
    if not isinstance(data0, np.ndarray):
        data_np = numpify(data, pad)  # still need original data for length-based bucketing
    else:
        data_np = data

    def get_bucket_probs(_buckets2instances):
        N = float(np.sum([len(ids) for ids in _buckets2instances.values()]))
        return {bid: len(ids) / N if N > 0. else 0. for bid, ids in _buckets2instances.items()}

    def shuffle_buckets(_buckets2instances):
        for bid in sorted(_buckets2instances.keys()):  # sorted: to keep deterministic
            rs.shuffle(_buckets2instances[bid])

    buckets2instances, _ = get_buckets(data, bucket_order, bucket_structure)
    n_buckets = len(buckets2instances)

    exact_epoch = True if len(data0) < n_buckets * batch_size else exact_epoch

    # if average instances/bucket smaller than batch_size: set exact_epoch = True
    # to avoid empty batches during debugging on small data samples

    def bucket_generator():
        buckets2instances, _ = get_buckets(data, bucket_order, bucket_structure)
        shuffle_buckets(buckets2instances)
        all_seen = False
        while not all_seen:
            bids, probs = zip(*sorted(get_bucket_probs(buckets2instances).items(), key=lambda x: x[0]))
            # sorted keys: to keep deterministic
            if np.sum(probs) == 0.:
                all_seen = True
            else:
                bid = rs.choice(bids, replace=False, p=probs)  # sample bucket according to remaining size
                batch_indices = buckets2instances[bid][:batch_size]
                buckets2instances[bid] = buckets2instances[bid][batch_size:]
                # if required by exact_epoch: also include last batch in bucket if too small
                if len(batch_indices) == batch_size or exact_epoch:
                    yield {k: data_np[k][batch_indices] for k in data_np}

    return GeneratorWithRestart(bucket_generator)


T = TypeVar('T')


def shuffle_and_batch(items: List[T], batch_size: int,
                      rng: Optional[random.Random] = None) \
        -> Iterator[List[T]]:
    """Optionally shuffles and batches items in a list.

    Args:
        - items: List of items to shuffle & batch.
        - batch_size: size of batches.
        - rng: random number generator if items should be shuffles, else None.

    Returns: Batch iterator
    """

    todo = list(range(len(items)))
    if rng is not None:
        rng.shuffle(todo)
    while todo:
        indices = todo[:batch_size]
        todo = todo[batch_size:]
        items_batch = [items[i] for i in indices]
        yield items_batch



"""

Weston et al. 2015
"Towards AI-Complete Question Answering: A Set of Prerequisite Toy Tasks"
arXiv preprint https://arxiv.org/abs/1502.05698.

Data: https://research.fb.com/projects/babi/
JTR download script: data/bAbI/download.sh

"""

import json
import argparse

# adapted from https://github.com/YerevaNN/Dynamic-memory-networks-in-Theano/
def load_babi_task(filename):
    """
    loads the babi data from its original format.
    """
    tasks = []
    task = None
    for i, line in enumerate(open(filename, 'r')):
        ID = int(line[0:line.find(' ')])

        # new data example
        if ID == 1:
            current_story = []
            counter = 0
            id_map = {}

        line = line.strip()
        line = line.replace('.', ' . ')
        line = line[line.find(' ')+1:]

        if '\t' not in line:
            # if line is not a question
            current_story.append(line)
            id_map[ID] = counter
            counter += 1
        else:
            # if the line is a question
            idx = line.find('?')
            tmp = line[idx+1:].split('\t')
            question = line[:idx]
            answer = tmp[1].strip()
            # copy by value.
            this_task = {"Story": [x for x in current_story], "Question": question, "Answer": answer}
            tasks.append(this_task)

    return tasks



def single_babi_example_in_JTR_format(instance):
    candidates_list = []
    qdict = {
        'question': instance['Question'],
        'candidates': [{'text': cand} for cand in candidates_list],
        'answers': [{'text': instance['Answer']}]
    }
    questions = [qdict]
    qset_dict = {
        'support': [{'text': supp} for supp in instance["Story"]],
        'questions': questions
    }

    return qset_dict



def convert_babi(path, n_instances=None):
    """
    Convert the babi data into jack format.

    Args:
        path: the file which should be converted
        n_instances: how many instances to filter out
    Returns: dictionary in jack format

    """
    # load data, select only first few if required.
    babi_data = load_babi_task(path)
    if n_instances != None:
        babi_data = babi_data[:n_instances]


    corpus = []
    for instance in babi_data:
        corpus.append(single_babi_example_in_JTR_format(instance))

    return {
        'meta': 'bAbI',
        'globals': {'candidates': []},
        'instances': corpus
    }


def main():
    """
        Main call function

    Usage:
        from command line:
            call with --help for help

    Returns: nothing
    """

    parser = argparse.ArgumentParser(description='The bAbI dataset to jack format converter.')
    parser.add_argument('infile',
                        help="path to the file to be converted (e.g. data/bAbI/tasks_1-20_v1-2/en/qa2_two-supporting-facts_train.txt)")
    parser.add_argument('outfile',
                        help="path to the jack format -generated output file (e.g. data/bAbI/jtr_format/train.jack.json)")
    parser.add_argument('-s', '--snippet', action="store_true",
                        help="Export a snippet (first 5 instances) instead of the full file")
    args = parser.parse_args()

    if args.snippet:
        jtr_corpus = convert_babi(args.infile, n_instances=4)
    else:
        jtr_corpus = convert_babi(args.infile)
    with open(args.outfile, 'w') as outfile:
        json.dump(jtr_corpus, outfile, indent=2)

if __name__ == "__main__":
    main()



import random
import re
from typing import List, Optional, Tuple

from jack.core.data_structures import QASetting, Answer
from jack.util import preprocessing
from jack.util.vocab import Vocab

__pattern = re.compile('\w+|[^\w\s]')


def tokenize(text):
    return __pattern.findall(text)


def token_to_char_offsets(text, tokenized_text):
    offsets = []
    offset = 0
    for t in tokenized_text:
        offset = text.index(t, offset)
        offsets.append(offset)
        offset += len(t)
    return offsets


def prepare_data(qa_setting: QASetting,
                 answers: Optional[List[Answer]],
                 vocab: Vocab,
                 lowercase: bool = False,
                 with_answers: bool = False,
                 wiq_contentword: bool = False,
                 spacy_nlp: bool = False,
                 max_support_length: int = None,
                 lemmatize=False,
                 with_lemmas=False) \
        -> Tuple[List[str], List[int], Optional[List[int]], int,
                 List[List[str]], List[List[int]], Optional[List[List[int]]], List[int],
                 List[List[float]], List[List[int]], List[List[Tuple[int, int]]]]:
    """Preprocesses a question and (optionally) answers:
    The steps include tokenization, lower-casing, translation to IDs,
    computing the word-in-question feature, computing token offsets,
    truncating supports, and computing answer spans.
    """
    supports = qa_setting.support
    question = qa_setting.question

    question_tokens, question_ids, question_length, question_lemmas, _ = preprocessing.nlp_preprocess(
        question, vocab, lowercase=lowercase, use_spacy=spacy_nlp,
        lemmatize=lemmatize, with_lemmas=with_lemmas, with_tokens_offsets=False)
    question_tokens_set = set(t.lower() for t in question_tokens)

    preprocessed_supports = [
        preprocessing.nlp_preprocess(
            support, vocab, lowercase=lowercase, use_spacy=spacy_nlp,
            lemmatize=lemmatize, with_lemmas=with_lemmas, with_tokens_offsets=True)
        for support in supports]

    all_support_tokens = [s[0] for s in preprocessed_supports]
    all_support_ids = [s[1] for s in preprocessed_supports]
    all_support_length = [s[2] for s in preprocessed_supports]
    all_support_lemmas = [s[3] for s in preprocessed_supports]
    all_token_offsets = [s[4] for s in preprocessed_supports]

    rng = random.Random(12345)

    all_word_in_question = []
    if with_lemmas:
        assert all_support_lemmas is not None
        for support_lemmas in all_support_lemmas:
            all_word_in_question.append([])
            if with_lemmas:
                for lemma in support_lemmas:
                    all_word_in_question[-1].append(float(
                        lemma in question_lemmas and (not wiq_contentword or (lemma.isalnum() and not lemma.is_stop))))
    else:
        for support_tokens in all_support_tokens:
            all_word_in_question.append([])
            for token in support_tokens:
                all_word_in_question[-1].append(
                    float(token.lower() in question_tokens_set and (not wiq_contentword or token.isalnum())))

    all_answer_spans = []
    for doc_idx, support_tokens in enumerate(all_support_tokens):
        min_answer = len(support_tokens)
        max_answer = 0
        token_offsets = all_token_offsets[doc_idx]

        answer_spans = []
        if with_answers:
            assert isinstance(answers, list)
            for a in answers:
                if a.doc_idx != doc_idx:
                    continue

                start = 0
                while start < len(token_offsets) and token_offsets[start] < a.span[0]:
                    start += 1

                if start == len(token_offsets):
                    continue

                end = start
                while end + 1 < len(token_offsets) and token_offsets[end + 1] < a.span[1]:
                    end += 1

                if (start, end) not in answer_spans:
                    answer_spans.append((start, end))
                    min_answer = min(min_answer, start)
                    max_answer = max(max_answer, end)

        # cut support whenever there is a maximum allowed length and recompute answer spans
        support_length = all_support_length[doc_idx]
        if max_support_length is not None and support_length > max_support_length > 0:
            if max_answer < max_support_length:
                # Find new start and end in the flattened support
                new_start = 0
                new_end = max_support_length
            else:
                offset = rng.randint(1, 11)
                new_end = max_answer
                new_start = max(0, min(min_answer, new_end + 2 * offset - max_support_length))
                while new_end - new_start > max_support_length - 2 * offset:
                    answer_spans = [(s, e) for s, e in answer_spans if e < new_end]
                    new_end = max(answer_spans, key=lambda span: span[1])[1]
                    new_start = max(0, min(min_answer, new_end + 2 * offset - max_support_length))
                new_end = min(new_end + offset, support_length)
                new_start = max(new_start - offset, 0)

            # Crop support according to new start and end pointers
            all_support_tokens[doc_idx] = support_tokens[new_start:new_end]
            all_support_ids[doc_idx] = all_support_ids[doc_idx][new_start:new_end]
            if with_lemmas:
                all_support_lemmas[doc_idx] = all_support_lemmas[doc_idx][new_start:new_end]
            answer_spans = [(s - new_start, e - new_start) for s, e in answer_spans]
            all_word_in_question[doc_idx] = all_word_in_question[doc_idx][new_start:new_end]
            all_support_length[doc_idx] = new_end - new_start
            all_token_offsets[doc_idx] = token_offsets[new_start:new_end]
        all_answer_spans.append(answer_spans)

    return question_tokens, question_ids, question_lemmas, question_length, \
           all_support_tokens, all_support_ids, all_support_lemmas, all_support_length, \
           all_word_in_question, all_token_offsets, all_answer_spans



import logging
import sys
from abc import abstractmethod
from typing import Mapping, List, Iterable, Tuple

import numpy as np
import torch
from torch import nn

from jack.core import ModelModule, SharedResources, TensorPort
from jack.core import reader
from jack.core.data_structures import Answer
from jack.core.data_structures import QASetting
from jack.core.tensorport import Ports

logger = reader.logger


class PyTorchModelModule(ModelModule):
    """This class represents an  abstract ModelModule for PyTorch models.

    It requires the implementation of 2 nn.modules that create predictions and the training outputs for the defined
    the ports.
    """

    def __init__(self, shared_resources: SharedResources):
        self.shared_resources = shared_resources
        # will be set in setup later
        self._prediction_module = None
        self._loss_module = None

    def __call__(self, batch: Mapping[TensorPort, np.ndarray],
                 goal_ports: List[TensorPort] = None) -> Mapping[TensorPort, np.ndarray]:
        """Runs a batch and returns values/outputs for specified goal ports.
        Args:
            batch: mapping from ports to values
            goal_ports: optional output ports, defaults to output_ports of this module will be returned

        Returns:
            A mapping from goal ports to tensors.
        """
        goal_ports = goal_ports or self.output_ports
        inputs = [p.create_torch_variable(batch.get(p), gpu=torch.cuda.device_count() > 0) for p in self.input_ports]
        outputs = self.prediction_module.forward(*inputs)
        ret = {p: p.torch_to_numpy(t) for p, t in zip(self.output_ports, outputs) if p in goal_ports}
        for p in goal_ports:
            if p not in ret and p in batch:
                ret[p] = batch[p]
        return ret

    @abstractmethod
    def create_prediction_module(self, shared_resources: SharedResources) -> nn.Module:
        """Creates and returns a PyTorch nn.Module for computing predictions.

        It takes inputs as defined by `input_ports` and produces  outputs as defined by `output_ports`"""
        raise NotImplementedError

    @abstractmethod
    def create_loss_module(self, shared_resources: SharedResources) -> nn.Module:
        """Creates and returns a PyTorch nn.Module for computing output necessary for training, such as a loss.

        It takes inputs as defined by `training_input_ports` and produces outputs as defined by
        `training_output_ports`."""
        raise NotImplementedError

    @property
    def prediction_module(self) -> nn.Module:
        return self._prediction_module

    @property
    def loss_module(self) -> nn.Module:
        return self._loss_module

    def setup(self, is_training=True):
        """Sets up the module.

        This usually involves creating the actual tensorflow graph. It is expected to be called after the input module
        is set up and shared resources, such as the vocab, config, etc., are prepared already at this point.
        """
        self._prediction_module = self.create_prediction_module(self.shared_resources)
        self._loss_module = self.create_loss_module(self.shared_resources)
        if torch.cuda.device_count() > 0:
            self._prediction_module.cuda()
            self._loss_module.cuda()

    def store(self, path):
        with open(path, 'wb') as f:
            torch.save({'prediction_module': self.prediction_module.state_dict(),
                        'loss_module': self.loss_module.state_dict()}, f)

    def load(self, path):
        with open(path, 'rb') as f:
            d = torch.load(f)
        self.prediction_module.load_state_dict(d['prediction_module'])
        self.loss_module.load_state_dict(d['loss_module'])


class PyTorchReader(reader.JTReader):
    """Tensorflow implementation of JTReader.

    A tensorflow reader reads inputs consisting of questions, supports and possibly candidates, and produces answers.
    It consists of three layers: input to tensor (input_module), tensor to tensor (model_module), and tensor to answer
    (output_model). These layers are called in-turn on a given input (list).
    """

    @property
    def model_module(self) -> PyTorchModelModule:
        return super().model_module

    def train(self, optimizer,
              training_set: Iterable[Tuple[QASetting, List[Answer]]],
              batch_size: int, max_epochs=10, hooks=tuple(), **kwargs):
        """This method trains the reader (and changes its state).

        Args:
            optimizer: optimizer
            training_set: the training instances.
            batch_size: size of training batches
            max_epochs: maximum number of epochs
            hooks: TrainingHook implementations that are called after epochs and batches
        """
        logger.info("Setting up data and model...")
        if not self._is_setup:
            # First setup shared resources, e.g., vocabulary. This depends on the input module.
            self.setup_from_data(training_set, is_training=True)
        batches = self.input_module.batch_generator(training_set, batch_size, is_eval=False)
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
        loss_idx = self.model_module.training_output_ports.index(Ports.loss)

        logger.info("Start training...")
        p_module = self.model_module.prediction_module
        l_module = self.model_module.loss_module
        for i in range(1, max_epochs + 1):
            for j, batch in enumerate(batches):
                for p, v in batch.items():
                    if isinstance(p, TensorPort):
                        batch[p] = p.create_torch_variable(v, gpu=torch.cuda.device_count() > 0)

                # zero the parameter gradients
                optimizer.zero_grad()
                pred_outputs = p_module.forward(
                    *(batch[p] for p in self.model_module.input_ports))
                batch.update(zip(self.model_module.output_ports, pred_outputs))
                train_outputs = l_module.forward(
                    *(batch[p] for p in self.model_module.training_input_ports))
                current_loss = train_outputs[loss_idx]
                current_loss.backward()
                optimizer.step()

                for hook in hooks:
                    hook.at_iteration_end(i, current_loss.data[0], set_name='train')

            # calling post-epoch hooks
            for hook in hooks:
                hook.at_epoch_end(i)



import torch
from torch.autograd import Function


def mask_for_lengths(length, max_length=None, mask_right=True, value=-1e6):
    max_length = max_length or length.max().data[0]
    mask = torch.cuda.IntTensor() if length.is_cuda else torch.IntTensor()
    mask = torch.arange(0, max_length, 1, out=mask)
    mask = torch.autograd.Variable(mask).type_as(length)
    mask /= length.unsqueeze(1)
    mask = mask.clamp(0, 1)
    mask = mask.float()
    if not mask_right:
        mask = 1.0 - mask
    mask *= value
    return mask


def segment_max(inputs, segment_ids, num_segments=None, default=0.0):
    # highly optimized to decrease the amount of actual invocation of pytorch calls
    # assumes that most segments have 1 or 0 elements
    segment_ids, indices = torch.sort(segment_ids)
    inputs = torch.index_select(inputs, 0, indices)
    output = SegmentMax.apply(inputs, segment_ids, num_segments, default)
    return output


class SegmentMax(Function):
    @staticmethod
    def forward(ctx, inputs, segment_ids, num_segments=None, default=0.0):
        # segments must be sorted by ids
        # highly optimized code, do not change if you don't know what you are doing.
        num_segments = num_segments or segment_ids.max() + 1

        zero_t = inputs[0].clone().view(1, -1)
        zero_t.fill_(default)

        if torch.is_tensor(segment_ids):
            segment_ids = segment_ids.cpu()

        lengths = []
        segm_lengths = []
        num_lengths = [0] * 10
        num_zeros = [0]

        lengths_extend = lengths.extend
        num_lengths_extend = num_lengths.extend
        segm_lengths_append = segm_lengths.append
        segm_lengths_extend = segm_lengths.extend

        _lengths = [[l] * l for l in range(len(num_lengths))]
        _zeros = [0] * num_segments

        def add_length(l):
            if l >= len(num_lengths):
                diff = l + 5 - len(num_lengths)
                num_lengths_extend([0] * diff)
                _lengths.extend([new_l] * new_l for new_l in range(len(_lengths), l + 5))
            num_lengths[l] += l
            segm_lengths_append(l)
            lengths_extend(_lengths[l])

        def add_zeros(n):
            segm_lengths_extend(_zeros[:n])
            num_zeros[0] += n

        offset = 0
        prev_s = segment_ids[0]
        if prev_s:
            add_zeros(prev_s)
        for i, s in enumerate(segment_ids):
            if prev_s != s:
                n_z = s - prev_s - 1
                if n_z > 0:
                    add_zeros(n_z)
                add_length(i - offset)
                offset = i
            prev_s = s
        add_length(segment_ids.shape[0] - offset)
        add_zeros(num_segments - prev_s - 1)

        lengths = torch.cuda.LongTensor(lengths) if inputs.is_cuda else torch.LongTensor(lengths)
        segm_lengths = torch.cuda.LongTensor(segm_lengths) if inputs.is_cuda else torch.LongTensor(segm_lengths)
        _, lengths_sorted = torch.sort(lengths, 0)
        _, segm_lengths_sorted = torch.sort(segm_lengths, 0)

        inputs_sorted = torch.index_select(inputs, 0, lengths_sorted)

        offset = [0]
        ctx.maxes = dict()

        def compute_segment(l, n):
            segment = inputs_sorted.narrow(0, offset[0], n)
            if l > 1:
                segment = segment.view(n // l, l, -1)
                ctx.maxes[l] = MyMax()
                segment = MyMax.forward(ctx.maxes[l], segment, dim=1)[0]
            offset[0] += n
            return segment

        segments = [compute_segment(l, n) for l, n in enumerate(num_lengths) if n > 0]
        segments = [zero_t.expand(num_zeros[0], zero_t.shape[1])] + segments

        segments = torch.cat(segments, 0)
        _, rev_segm_sorted = torch.sort(segm_lengths_sorted)

        ctx.rev_segm_sorted = rev_segm_sorted
        ctx.num_lengths = num_lengths
        ctx.lengths_sorted = lengths_sorted
        ctx.num_zeros = num_zeros[0]

        output = torch.index_select(segments, 0, rev_segm_sorted)

        return output

    @staticmethod
    def backward(ctx, grad_outputs):
        size = grad_outputs.size(1)
        segm_sorted = torch.sort(ctx.rev_segm_sorted)[1]
        grad_outputs = torch.index_select(grad_outputs, 0, segm_sorted)

        offset = [ctx.num_zeros]

        def backward_segment(l, n):
            segment_grad = grad_outputs.narrow(0, offset[0], n // l)
            if l > 1:
                segment_grad = MyMax.backward(ctx.maxes[l], segment_grad)[0].view(n, size)
            offset[0] += n // l
            return segment_grad

        segment_grads = [backward_segment(l, n) for l, n in enumerate(ctx.num_lengths) if n > 0]
        grads = torch.cat(segment_grads, 0)
        rev_length_sorted = torch.sort(ctx.lengths_sorted)[1]
        grads = torch.index_select(grads, 0, rev_length_sorted)

        return grads, None, None, None


class MyMax:
    @classmethod
    def forward(cls, ctx, input, dim, keepdim=None):
        ctx.dim = dim
        ctx.keepdim = False if keepdim is None else keepdim
        ctx.input_size = input.size()

        output, indices = input.max(dim=dim, keepdim=ctx.keepdim)
        ctx.indices = indices

        return output, indices

    @classmethod
    def backward(cls, ctx, grad_output, grad_indices=None):
        grad_input = torch.autograd.Variable(grad_output.data.new(*ctx.input_size).zero_())
        dim = ctx.dim
        indices = ctx.indices
        if not ctx.keepdim:
            grad_output = grad_output.unsqueeze(dim)
            indices = indices.unsqueeze(dim)

        grad_input.scatter_(dim, indices, grad_output)

        return grad_input, None, None



# -*- coding: utf-8 -*-

import tensorflow as tf

from jack.util.tf import misc


def conv_char_embedding(num_chars, repr_dim, word_chars, word_lengths, word_sequences=None,
                        conv_width=5, emb_initializer=tf.random_normal_initializer(0.0, 0.1), scope=None):
    """Build simple convolutional character based embeddings for words with a fixed filter and size.

    After the convolution max-pooling over characters is employed for each filter. If word sequences are given,
    these will be embedded with the newly created embeddings.
    """
    # "fixed PADDING on character level"
    pad = tf.zeros(tf.stack([tf.shape(word_lengths)[0], conv_width // 2]), tf.int32)
    word_chars = tf.concat([pad, word_chars, pad], 1)

    with tf.variable_scope(scope or "char_embeddings"):
        char_embedding_matrix = \
            tf.get_variable("char_embedding_matrix", shape=(num_chars, repr_dim),
                            initializer=emb_initializer, trainable=True)

        max_word_length = tf.reduce_max(word_lengths)
        embedded_chars = tf.nn.embedding_lookup(char_embedding_matrix, tf.cast(word_chars, tf.int32))

        with tf.variable_scope("conv"):
            # create filter like this to get fan-in and fan-out right for initializers depending on those
            filter = tf.get_variable("filter", [conv_width * repr_dim, repr_dim])
            filter_reshaped = tf.reshape(filter, [conv_width, repr_dim, repr_dim])
            # [B, T, S + pad_right]
            conv_out = tf.nn.conv1d(embedded_chars, filter_reshaped, 1, "VALID")
            conv_mask = tf.expand_dims(misc.mask_for_lengths(word_lengths, max_length=max_word_length), 2)
            conv_out = conv_out + conv_mask

        embedded_words = tf.reduce_max(conv_out, 1)

    if word_sequences is None:
        return embedded_words

    if not isinstance(word_sequences, list):
        word_sequences = [word_sequences]
    all_embedded = []
    for word_idxs in word_sequences:
        all_embedded.append(tf.nn.embedding_lookup(embedded_words, word_idxs))

    return all_embedded


def conv_char_embedding_multi_filter(
        num_chars, filter_sizes, embedding_size, word_chars, word_lengths, word_sequences=None,
        emb_initializer=tf.random_normal_initializer(0.0, 0.1), projection_size=None, scope=None):
    """Build convolutional character based embeddings for words with multiple filters.

    Filter sizes is a list and each the position of each size in the list entry refers to its corresponding conv width.
    It can also be 0 (i.e., no filter of that conv width). E.g., sizes [4, 0, 7, 8] will create 4 conv filters of width
    1, no filter of width 2, 7 of width 3 and 8 of width 4. After the convolution max-pooling over characters is
    employed for each filter.

    embedding_size refers to the size of the character embeddings and projection size, if given, to the final size of
    the embedded characters after a final projection. If it is None, no projection will be applied and the resulting
    size is the sum of all filters.

    If word sequences are given, these will be embedded with the newly created embeddings.
    """
    with tf.variable_scope(scope or "char_embeddings"):
        char_embedding_matrix = \
            tf.get_variable("char_embedding_matrix", shape=(num_chars, embedding_size),
                            initializer=emb_initializer, trainable=True)

        pad = tf.zeros(tf.stack([tf.shape(word_lengths)[0], len(filter_sizes) // 2]), tf.int32)
        word_chars = tf.concat([pad, word_chars, pad], 1)

        max_word_length = tf.reduce_max(word_lengths)
        embedded_chars = tf.nn.embedding_lookup(char_embedding_matrix, tf.cast(word_chars, tf.int32))
        conv_mask = tf.expand_dims(misc.mask_for_lengths(word_lengths, max_length=max_word_length), 2)

        embedded_words = []
        for i, size in enumerate(filter_sizes):
            if size == 0:
                continue
            conv_width = i + 1
            with tf.variable_scope("conv_%d" % conv_width):
                # create filter like this to get fan-in and fan-out right for initializers depending on those
                filter = tf.get_variable("filter", [conv_width * embedding_size, size])
                filter_reshaped = tf.reshape(filter, [conv_width, embedding_size, size])
                cut = len(filter_sizes) // 2 - conv_width // 2
                embedded_chars_conv = embedded_chars[:, cut:-cut, :] if cut else embedded_chars
                conv_out = tf.nn.conv1d(embedded_chars_conv, filter_reshaped, 1, "VALID")
                conv_out += conv_mask
                embedded_words.append(tf.reduce_max(conv_out, 1))

        embedded_words = tf.concat(embedded_words, 1)
        if projection_size is not None:
            embedded_words = tf.layers.dense(embedded_words, projection_size)

    if word_sequences is None:
        return embedded_words

    if not isinstance(word_sequences, list):
        word_sequences = [word_sequences]
    all_embedded = []
    for word_idxs in word_sequences:
        embedded_words = tf.nn.embedding_lookup(embedded_words, word_idxs)
        all_embedded.append(embedded_words)

    return all_embedded



import tensorflow as tf

from jack.readers.classification.shared import AbstractSingleSupportClassificationModel
from jack.readers.natural_language_inference import prediction_layer
from jack.util.tf.embedding import conv_char_embedding
from jack.util.tf.modular_encoder import modular_encoder


class ModularNLIModel(AbstractSingleSupportClassificationModel):
    def forward_pass(self, shared_resources, embedded_question, embedded_support, num_classes, tensors):
        model = shared_resources.config['model']
        repr_dim = shared_resources.config['repr_dim']
        dropout = shared_resources.config.get("dropout")

        if shared_resources.config.get('with_char_embeddings'):
            [char_emb_question, char_emb_support] = conv_char_embedding(
                len(shared_resources.char_vocab), shared_resources.config['repr_dim'], tensors.word_chars,
                tensors.word_char_length, [tensors.question_batch_words, tensors.support_batch_words])
            inputs = {'hypothesis': embedded_question, 'premise': embedded_support,
                      'char_hypothesis': char_emb_question, 'char_premise': char_emb_support}
            inputs_length = {'hypothesis': tensors.question_length, 'premise': tensors.support_length,
                             'char_hypothesis': tensors.question_length, 'char_premise': tensors.support_length}
        else:
            inputs = {'hypothesis': embedded_question, 'premise': embedded_support}
            inputs_length = {'hypothesis': tensors.question_length, 'premise': tensors.support_length}

        if dropout:
            for k in inputs:
                inputs[k] = tf.cond(tensors.is_eval, lambda: inputs[k], lambda: tf.nn.dropout(inputs[k], 1.0 - dropout))

        inputs_mapping = {}

        encoder_config = model['encoder_layer']
        encoded, _, _ = modular_encoder(
            encoder_config, inputs, inputs_length, inputs_mapping, repr_dim, dropout, tensors.is_eval)

        with tf.variable_scope('prediction_layer'):
            prediction_layer_config = model['prediction_layer']
            encoded_question = encoded[prediction_layer_config.get('hypothesis', 'hypothesis')]
            encoded_support = encoded[prediction_layer_config.get('premise', 'premise')]

            if 'repr_dim' not in prediction_layer_config:
                prediction_layer_config['repr_dim'] = repr_dim
            prediction_layer_config['dropout'] = dropout if prediction_layer_config.get('dropout', False) else 0.0
            logits = prediction_layer.prediction_layer(
                encoded_question, tensors.question_length, encoded_support,
                tensors.support_length, num_classes, tensors.is_eval, **prediction_layer_config)

        return logits



"""
This files allows creating a jtr datafile for the SNLI corpus,
whereby each instance receives support under the form of 'related' instances
"""


import json

__candidate_labels = ['entailment', 'neutral', 'contradiction']
__candidates = [{'text': cl} for cl in __candidate_labels]


def convert_snli(snli_file_jsonl):
    """ io SNLI files into jack format.
    Data source: http://nlp.stanford.edu/projects/snli/snli_1.0.zip
    Files to be converted: snli_1.0_dev.jsonl, snli_1.0_train.jsonl, snli_1.0_test.jsonl
    (the *.txt files contain the same data in a different format)

    Format:
        - support = the premise = 'sentence1' in original SNLI data (as id, use 'captionID', the id of sentence1)
        - question = the hypothesis = 'sentence2' in original SNLI data
    Notes:
        - instances with gold labels '-' are removed from the corpus
    """
    with open(snli_file_jsonl, 'r') as f:
        data = [__convert_snli_instance(json.loads(line.strip())) for line in f.readlines()]

        return {'meta': 'SNLI',
                'globals': {'candidates': __candidates},
                'instances': [d for d in data if d]  # filter out invalid ones
                }


def __convert_snli_instance(instance):
    queb = {}
    if instance['gold_label'] in __candidate_labels:
        queb['id'] = instance['pairID']
        queb['support'] = [
            {'id': instance.get('captionID'), 'text': instance['sentence1']}]
        queb['questions'] = [
            {'question': instance['sentence2'],
             'answers': [{'text': instance['gold_label']}]}]
    return queb


def main():
    import sys
    if len(sys.argv) == 2:
        corpus = convert_snli(sys.argv[1])
    else:
        for corpus_name in ["dev", "train", "test"]:
            corpus = convert_snli("./data/SNLI/snli_1.0/snli_1.0_%s.jsonl" % corpus_name)
            with open("./data/SNLI/snli_1.0/snli_1.0_%s_jtr_v1.json" % corpus_name, 'w') as outfile:
                print("Create file snli_1.0_%s_jtr.json" % corpus_name)
                json.dump(corpus, outfile, indent=2)

        # create train set test data
        corpus = convert_snli("./data/SNLI/snli_1.0/snli_1.0_train.jsonl")
        corpus['instances'] = corpus['instances'][:2000]
        with open("./tests/test_data/SNLI/2000_samples_train_jtr_v1.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        corpus['instances'] = corpus['instances'][:100]
        with open("./tests/test_data/SNLI/overfit.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        # create snippets and overfit test data
        corpus['instances'] = corpus['instances'][:10]
        with open("./data/SNLI/snli_1.0/snli_1.0_debug_jtr_v1.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)
        with open("./data/SNLI/snippet.jtr_v1.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        # create dev set test data
        corpus = convert_snli("./data/SNLI/snli_1.0/snli_1.0_dev.jsonl")
        corpus['instances'] = corpus['instances'][:1000]
        with open("./tests/test_data/SNLI/1000_samples_dev_jtr_v1.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

        # create dev set test data
        corpus = convert_snli("./data/SNLI/snli_1.0/snli_1.0_test.jsonl")
        corpus['instances'] = corpus['instances'][:2000]
        with open("./tests/test_data/SNLI/2000_samples_test_jtr_v1.json", 'w') as outfile:
            json.dump(corpus, outfile, indent=2)

if __name__ == "__main__":
    main()



"""
This files allows creating a jtr train and test datafiles for the NYT corpus
based on jtr/data/NYT/naacl2013.txt
* train file: one jtr per train instance ; candidates = all entity pairs
* test file: one jtr per test relation, with all correct answers in the valid answers list

usage:
python3 NYT2jtr path/to/naacl2013.txt mode >  naacl2013_mode.jtr.json
in which mode = train or test
"""

import json


def load_naacl2013(path, mode):
    assert mode in ['train', 'test', 'Train', 'Test'], 'provide mode argument: train or test'
    facts = []
    # load facts of type mode
    with open(path) as fID:
        for line in fID:
            if len(line.strip()) > 0:
                r, e1, e2, typ, truth = line.strip().split('\t')
                rel = r
                truth = {'1.0': True, '0.0': False}[truth]
                tup = '(%s|||%s)' % (e1, e2)
                if typ.lower() == mode.lower():
                    facts.append(((rel, tup), truth, typ))

    # global candidates: all entity tuples appearing in the training data
    # (for model F, test data should have no tuples not occurring in training data)
    tup_candidates_train = sorted(list(set([f[0][1] for f in facts if f[2].lower() == 'train'])))
    tup_candidates_test = sorted(list(set([f[0][1] for f in facts if f[2].lower() == 'test'])))

    if mode.lower() == 'train':
        return create_train_jtr(facts, tup_candidates_train)
    elif mode.lower() == 'test':
        return create_test_jtr(facts, tup_candidates_test)


def create_train_jtr(trainfacts, tuples):
    """
    TODO comment
    one jack per positive train fact

    Args:
        trainfacts:
        tuples:

    Returns:

    """

    instances = [{
                     'support': [],
                     'questions': [{
                         'question': fact[0][0],
                         'candidates': '#/globals/candidates',
                         'answers': [{
                             'text': fact[0][1]
                         }]
                     }]
                     # only add true facts for training (data should be consistent with that)
                  } for fact in trainfacts if fact[1]]

    # @Johannes: originally we had 'instances' as the jack format;
    # now: added metadata field 'meta' and 'globals' field with the overall candidates.
    return {
        'meta': 'MFmodel-train',
        'globals': {
            'candidates': [{'text': tup} for tup in tuples]},
        'instances': instances
    }


def create_test_jtr(testfacts, tuples):
    """
    TODO comment
    one jack with all correct answers per relation in the test data

    Args:
        testfacts:
        tuples:

    Returns:

    """
    # map test relations to all true/false entity tuples
    relmap = {}
    for fact in testfacts:
        rel, tup = fact[0]
        truth = fact[1]
        if not truth:
            continue
        if rel not in relmap:
            relmap[rel] = {
                'candidates': set(),
                'answers': set()
            }

        # relmap[rel]['candidates'].add(tup)
        relmap[rel]['answers'].add(tup)

    # test instances:
    instances = [{
                     'support': [],
                     'questions': [{
                         'question': rel,
                         'candidates': [{'text': c} for c in relmap[rel]['candidates']],
                         'answers': [{'text': a} for a in relmap[rel]['answers']]
                     }]
                  } for rel in relmap]
    return {
        'meta': 'MFmodel-test',
        'globals': {
            'candidates': [{'text': tup} for tup in tuples]},
        'instances': instances
          }


def main():
    import sys
    if len(sys.argv) == 4:
        data = load_naacl2013(sys.argv[1], sys.argv[2])
        with open(sys.argv[3], 'w') as outfile:
            json.dumps(data, outfile, indent=2)

    else:
        print("""Usage:
    python3 NYT2jtr.py path/to/naacl2013 {mode} /save/to/naacl2013_mode.jack.json
        where {mode} = {train, test}""")

if __name__ == "__main__":
    main()



"""

jtr converter for the fb15k dataset.

Bordes, Antoine, et al.
"Translating embeddings for modeling multi-relational data."
Advances in neural information processing systems. 2013.

Original paper:
        https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data
Data:   https://everest.hds.utc.fr/lib/exe/fetch.php?media=en:fb15k.tgz
Web:    https://everest.hds.utc.fr/doku.php?id=en:transe
JTR download script: data/FB15k/download.sh

Metadata:

Training data:
    483142 triples (subject, relation, object)
    14951 different entities
    1345 different relation types

"""

import argparse
import gc
import json
from collections import defaultdict


def load_fb15k_triples(path):
    """
    Loads the raw data from file provided.

    Args:
        path: path to the file

    Returns: triples
    """
    with open(path, 'r') as f:
        triples = [line.strip('\n').split('\t') for line in f.readlines()]
    return triples


def extract_unique_entities_and_relations(triples):
    """
    Identifies unique entities and relation types in collection of triples.

    Args:
        triples: List of string triples.

    Returns:
        unique_entities: List of strings
        unique_relations: List of strings
    """
    s_entities = set([triple[0] for triple in triples])
    o_entities = set([triple[2] for triple in triples])
    r_types = set([triple[1] for triple in triples])

    unique_relations = sorted(list(r_types))
    unique_entities = sorted(list(s_entities | o_entities))  # union of sets

    return unique_entities, unique_relations


def get_facts_per_entity(triples):
    """
    Obtain dictionary with all train fact ids that contain an entity.

    Args:
        triples: List of fact triples

    Returns:
        Dictionary entity --> fact IDs it participates in
    """
    d = defaultdict(set)
    for i_triple, triple in enumerate(triples):
        d[triple[0]].add(i_triple)
        d[triple[2]].add(i_triple)
    return d


def get_facts_per_relation(triples):
    """
    Obtain dictionary with all train fact ids that contain a relation type.

    Args:
        triples: List of fact triples

    Returns:
        Dictionary relation type --> fact IDs it participates in
    """
    d = defaultdict(set)
    for i_triple, triple in enumerate(triples):
        d[triple[1]].add(i_triple)
    return d


def get_fact_neighbourhoods(triples, facts_per_entity, facts_per_relation,
                            include_relations=False):
    """
    Extracts neighbouring facts for a collection of triples. neighbouring
    facts of fact f are such facts that share at least an entity with f.
    If relations are included, facts which share a relation are also considered
    neighbours.

    Args:
        triples: list of facts triples
        facts_per_entity: dictionary; The facts an entity appears in
        facts_per_relation: dictionary; The facts a relation appears in
        include_relations: boolean. whether facts sharing the relation should
            be considered neighbours as well.

    Returns:
        fact_neighbourhoods: dictionary mapping fact ID to set of fact IDs.
    """
    fact_neighbourhoods = defaultdict(set)
    for i_triple, triple in enumerate(triples):
        # get triple ids which share subject, object or rel. with current triple
        subject_neighbours = facts_per_entity[triple[0]]
        object_neighbours = facts_per_entity[triple[2]]
        relation_neighbours = set()
        if include_relations:
            relation_neighbours = facts_per_relation[triple[1]]

        fact_neighbourhoods[i_triple].update(subject_neighbours)
        fact_neighbourhoods[i_triple].update(object_neighbours)
        fact_neighbourhoods[i_triple].update(relation_neighbours)

    return fact_neighbourhoods


def convert_fb15k(triples, neighbourhoods):
    """
    Converts into jack format.
    Args:
        triples: fact triples that should be converted.
        neighbourhoods: dictionary of supporting facts per triple

    Returns:
        jack formatted fb15k data.
    """
    instances = []
    for i, triple in enumerate(triples):
        if not i % 1000:
            # print(i)
            gc.collect()

        # obtain supporting facts for this triple
        neighbour_ids = neighbourhoods.get(i)
        qset_dict = {}
        if neighbour_ids:
            neighbour_triples = [triples[ID] for ID in neighbour_ids]
            qset_dict['support'] = [" ".join(t) for t in neighbour_triples]

        qset_dict['questions'] = [{
            "question": " ".join(triple),
            "answers": ["True"]
        }]
        instances.append(qset_dict)

    return {
        'meta': 'FB15K-237 dataset.',
        'instances': instances
    }


def main():
    parser = argparse.ArgumentParser(description='FB15K to jack format converter.')
    #
    parser.add_argument('infile',
                        help="dataset path you're interested in, train/dev/test."
                             "(e.g. data/FB15k-237/Release/train.txt)")
    parser.add_argument('outfile',
                        help="path to the jack format -generated output file (e.g. data/FB15K-237/FB15k_train.jack.json)")
    # parser.add_argument('dataset', choices=['cnn', 'dailymail'],
    #                     help="which dataset to access: cnn or dailymail")
    parser.add_argument('--support', default='',
                        help="use training set path here (e.g. data/FB15k-237/Release/train.txt)."
                             "Default is not to create supporting facts.")
    args = parser.parse_args()

    print("Loading data...")
    # load data from files into fact triples
    triples = load_fb15k_triples(args.infile)

    # get neighbouring facts for each fact in triples
    if args.support:
        print("Creating fact neighbourhoods as support...")
        if args.infile == args.support:
            reference_triples = triples
        else:
            reference_triples = load_fb15k_triples(args.support)
        facts_per_entity = get_facts_per_entity(reference_triples)
        facts_per_relation = get_facts_per_relation(reference_triples)
        neighbourhoods = get_fact_neighbourhoods(triples, facts_per_entity, facts_per_relation)
    else:
        neighbourhoods = dict()

    # dump the entity and relation ids for understanding the jack contents.
    print("Convert to json...")
    corpus = convert_fb15k(triples, neighbourhoods)
    with open(args.outfile, 'w') as outfile:
        json.dump(corpus, outfile, indent=2)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import json
import os

import numpy as np

from jack.io.embeddings import Embeddings


def load_memory_map_dir(directory: str) -> Embeddings:
    """
    Loads embeddings from a memory map directory to allow lazy loading (and reduce the memory usage).
    Args:
        directory: a file prefix. This function loads two files in the directory: a meta json file with shape information
        and the vocabulary, and the actual memory map file.

    Returns:
        Embeddings object with a lookup matrix that is backed by a memory map.

    """
    meta_file = os.path.join(directory, "meta.json")
    mem_map_file = os.path.join(directory, "memory_map")
    with open(meta_file, "r") as f:
        meta = json.load(f)
    shape = tuple(meta['shape'])
    vocab = meta['vocab']
    mem_map = np.memmap(mem_map_file, dtype='float32', mode='r+', shape=shape)
    result = Embeddings(vocab, mem_map, filename=directory, emb_format="memory_map_dir")
    return result


def save_as_memory_map_dir(directory: str, emb: Embeddings):
    """
    Saves the given embeddings as memory map file and corresponding meta data in a directory.
    Args:
        directory: the directory to store the memory map file in (called `memory_map`) and the meta file (called
        `meta.json` that stores the shape of the memory map and the actual vocabulary.
        emb: the embeddings to store.
    """
    if not os.path.exists(directory):
        os.makedirs(directory)

    meta_file = os.path.join(directory, "meta.json")
    mem_map_file = os.path.join(directory, "memory_map")
    with open(meta_file, "w") as f:
        json.dump({
            "vocab": emb.vocabulary,
            "shape": emb.shape
        }, f)
    mem_map = np.memmap(mem_map_file, dtype='float32', mode='w+', shape=emb.shape)
    mem_map[:] = emb.lookup[:]
    mem_map.flush()
    del mem_map



# -*- coding: utf-8 -*-

import torch
import torch.nn.functional as F
from torch import nn


class XQAMinCrossentropyLossModule(nn.Module):
    def forward(self, start_scores, end_scores, answer_span, answer_to_question):
        """very common XQA loss function."""
        long_tensor = torch.cuda.LongTensor if torch.cuda.device_count() > 0 else torch.LongTensor
        answer_span = answer_span.type(long_tensor)
        start, end = answer_span[:, 0], answer_span[:, 1]

        batch_size1 = start.data.shape[0]
        batch_size2 = start_scores.data.shape[0]
        is_aligned = batch_size1 == batch_size2

        start_scores = start_scores if is_aligned else torch.index_select(start_scores, dim=0, index=answer_to_question)
        end_scores = end_scores if is_aligned else torch.index_select(end_scores, dim=0, index=answer_to_question)

        partitioned_loss = []
        for i, j in enumerate(answer_to_question):
            j = j.data[0]
            while j >= len(partitioned_loss):
                partitioned_loss.append([])
            loss = -torch.index_select(F.log_softmax(start_scores[i], dim=0), dim=0, index=start[i])
            loss -= torch.index_select(F.log_softmax(end_scores[i], dim=0), dim=0, index=end[i])
            partitioned_loss[j].append(loss)

        for j, l in enumerate(partitioned_loss):
            partitioned_loss[j] = torch.stack(l).min()

        loss = torch.stack(partitioned_loss).mean()
        return loss



import tensorflow as tf

from jack.util.tf import attention, sequence_encoder, misc
from jack.util.tf.segment import segment_top_k, segment_softmax


def answer_layer(encoded_question, question_length, encoded_support, support_length,
                 support2question, answer2support, is_eval, correct_start=None, topk=1, max_span_size=10000,
                 encoder=None, module='bilinear', repr_dim=100, **kwargs):
    if module == 'bilinear':
        return bilinear_answer_layer(
            repr_dim, encoded_question, question_length, encoded_support, support_length,
            support2question, answer2support, is_eval, topk, max_span_size)
    elif module == 'mlp':
        return mlp_answer_layer(repr_dim, encoded_question, question_length, encoded_support, support_length,
                                support2question, answer2support, is_eval, topk, max_span_size)
    elif module == 'conditional':
        return conditional_answer_layer(
            repr_dim, encoded_question, question_length, encoded_support, support_length,
            correct_start, support2question, answer2support, is_eval, topk, max_span_size)
    elif module == 'conditional_bilinear':
        return conditional_answer_layer(
            repr_dim, encoded_question, question_length, encoded_support, support_length,
            correct_start, support2question, answer2support, is_eval, topk, max_span_size, bilinear=True)
    elif module == 'san':
        return san_answer_layer(
            repr_dim, encoded_question, question_length, encoded_support, support_length,
            support2question, answer2support, is_eval, topk, max_span_size, **kwargs)
    elif module == 'bidaf':
        if 'repr_dim' not in encoder:
            encoder['repr_dim'] = repr_dim
        encoded_support_end = sequence_encoder.encoder(
            encoded_support, support_length, name='encoded_support_end', is_eval=is_eval, **encoder)
        encoded_support_end = tf.concat([encoded_support, encoded_support_end], 2)
        return bidaf_answer_layer(encoded_support, encoded_support_end, support_length,
                                  support2question, answer2support, is_eval, topk=1, max_span_size=10000)
    else:
        raise ValueError("Unknown answer layer type: %s" % module)


def compute_question_state(encoded_question, question_length):
    attention_scores = tf.layers.dense(encoded_question, 1, name="question_attention")
    q_mask = misc.mask_for_lengths(question_length)
    attention_scores = attention_scores + tf.expand_dims(q_mask, 2)
    question_attention_weights = tf.nn.softmax(attention_scores, 1, name="question_attention_weights")
    question_state = tf.reduce_sum(question_attention_weights * encoded_question, 1)
    return question_state


def bilinear_answer_layer(size, encoded_question, question_length, encoded_support, support_length,
                          support2question, answer2support, is_eval, topk=1, max_span_size=10000):
    """Answer layer for multiple paragraph QA."""
    # computing single time attention over question
    size = encoded_support.get_shape()[-1].value
    question_state = compute_question_state(encoded_question, question_length)

    # compute logits
    hidden = tf.gather(tf.layers.dense(question_state, 2 * size, name="hidden"), support2question)
    hidden_start, hidden_end = tf.split(hidden, 2, 1)

    support_mask = misc.mask_for_lengths(support_length)

    start_scores = tf.einsum('ik,ijk->ij', hidden_start, encoded_support)
    start_scores = start_scores + support_mask

    end_scores = tf.einsum('ik,ijk->ij', hidden_end, encoded_support)
    end_scores = end_scores + support_mask

    return compute_spans(start_scores, end_scores, answer2support, is_eval, support2question, topk, max_span_size)


def mlp_answer_layer(size, encoded_question, question_length, encoded_support, support_length,
                     support2question, answer2support, is_eval, topk=1, max_span_size=10000):
    """Answer layer for multiple paragraph QA."""
    # computing single time attention over question
    question_state = compute_question_state(encoded_question, question_length)

    # compute logits
    static_input = tf.concat([tf.gather(tf.expand_dims(question_state, 1), support2question) * encoded_support,
                              encoded_support], 2)

    hidden = tf.gather(tf.layers.dense(question_state, 2 * size, name="hidden_1"), support2question)
    hidden = tf.layers.dense(
        static_input, 2 * size, use_bias=False, name="hidden_2") + tf.expand_dims(hidden, 1)

    hidden_start, hidden_end = tf.split(tf.nn.relu(hidden), 2, 2)

    support_mask = misc.mask_for_lengths(support_length)

    start_scores = tf.layers.dense(hidden_start, 1, use_bias=False, name="start_scores")
    start_scores = tf.squeeze(start_scores, [2])
    start_scores = start_scores + support_mask

    end_scores = tf.layers.dense(hidden_end, 1, use_bias=False, name="end_scores")
    end_scores = tf.squeeze(end_scores, [2])
    end_scores = end_scores + support_mask

    return compute_spans(start_scores, end_scores, answer2support, is_eval, support2question, topk, max_span_size)


def compute_spans(start_scores, end_scores, answer2support, is_eval, support2question,
                  topk=1, max_span_size=10000, correct_start=None):
    max_support_length = tf.shape(start_scores)[1]
    _, _, num_doc_per_question = tf.unique_with_counts(support2question)
    offsets = tf.cumsum(num_doc_per_question, exclusive=True)
    doc_idx_for_support = tf.range(tf.shape(support2question)[0]) - tf.gather(offsets, support2question)

    def train():
        gathered_end_scores = tf.gather(end_scores, answer2support)
        gathered_start_scores = tf.gather(start_scores, answer2support)

        if correct_start is not None:
            # assuming we know the correct start we only consider ends after that
            left_mask = misc.mask_for_lengths(tf.cast(correct_start, tf.int32), max_support_length, mask_right=False)
            gathered_end_scores = gathered_end_scores + left_mask

        predicted_start_pointer = tf.argmax(gathered_start_scores, axis=1, output_type=tf.int32)
        predicted_end_pointer = tf.argmax(gathered_end_scores, axis=1, output_type=tf.int32)

        return (start_scores, end_scores,
                tf.gather(doc_idx_for_support, answer2support), predicted_start_pointer, predicted_end_pointer)

    def eval():
        # we collect spans for top k starts and top k ends and select the top k from those top 2k
        doc_idx1, start_pointer1, end_pointer1, span_score1 = _get_top_k(
            start_scores, end_scores, topk, max_span_size, support2question)
        doc_idx2, end_pointer2, start_pointer2, span_score2 = _get_top_k(
            end_scores, start_scores, topk, -max_span_size, support2question)

        doc_idx = tf.concat([doc_idx1, doc_idx2], 1)
        start_pointer = tf.concat([start_pointer1, start_pointer2], 1)
        end_pointer = tf.concat([end_pointer1, end_pointer2], 1)
        span_score = tf.concat([span_score1, span_score2], 1)

        _, idx = tf.nn.top_k(span_score, topk)

        r = tf.range(tf.shape(span_score)[0], dtype=tf.int32)
        r = tf.reshape(tf.tile(tf.expand_dims(r, 1), [1, topk]), [-1, 1])

        idx = tf.concat([r, tf.reshape(idx, [-1, 1])], 1)
        doc_idx = tf.gather_nd(doc_idx, idx)
        start_pointer = tf.gather_nd(start_pointer, idx)
        end_pointer = tf.gather_nd(end_pointer, idx)

        return start_scores, end_scores, tf.gather(doc_idx_for_support, doc_idx), start_pointer, end_pointer

    return tf.cond(is_eval, eval, train)


def _get_top_k(scores1, scores2, k, max_span_size, support2question):
    max_support_length = tf.shape(scores1)[1]
    doc_idx, pointer1, topk_scores1 = segment_top_k(scores1, support2question, k)

    # [num_questions * topk]
    doc_idx_flat = tf.reshape(doc_idx, [-1])
    pointer_flat1 = tf.reshape(pointer1, [-1])

    # [num_questions * topk, support_length]
    scores_gathered2 = tf.gather(scores2, doc_idx_flat)
    if max_span_size < 0:
        pointer_flat1, max_span_size = pointer_flat1 + max_span_size + 1, -max_span_size
    left_mask = misc.mask_for_lengths(tf.cast(pointer_flat1, tf.int32),
                                      max_support_length, mask_right=False)
    right_mask = misc.mask_for_lengths(tf.cast(pointer_flat1 + max_span_size, tf.int32),
                                       max_support_length)
    scores_gathered2 = scores_gathered2 + left_mask + right_mask

    pointer2 = tf.argmax(scores_gathered2, axis=1, output_type=tf.int32)

    topk_score2 = tf.gather_nd(scores2, tf.stack([doc_idx_flat, pointer2], 1))

    return doc_idx, pointer1, tf.reshape(pointer2, [-1, k]), topk_scores1 + tf.reshape(topk_score2, [-1, k])


def conditional_answer_layer(size, encoded_question, question_length, encoded_support, support_length,
                             correct_start, support2question, answer2support, is_eval, topk=1, max_span_size=10000,
                             bilinear=False):
    question_state = compute_question_state(encoded_question, question_length)
    question_state = tf.gather(question_state, support2question)

    # Prediction
    # start
    if bilinear:
        hidden_start = tf.layers.dense(question_state, size, name="hidden_start")
        start_scores = tf.einsum('ik,ijk->ij', hidden_start, encoded_support)
    else:
        static_input = tf.concat([tf.expand_dims(question_state, 1) * encoded_support, encoded_support], 2)
        hidden_start = tf.layers.dense(question_state, size, name="hidden_start_1")
        hidden_start = tf.layers.dense(
            static_input, size, use_bias=False, name="hidden_start_2") + tf.expand_dims(hidden_start, 1)
        start_scores = tf.layers.dense(tf.nn.relu(hidden_start), 1, use_bias=False, name="start_scores")
        start_scores = tf.squeeze(start_scores, [2])

    support_mask = misc.mask_for_lengths(support_length)
    start_scores = start_scores + support_mask

    max_support_length = tf.shape(start_scores)[1]
    _, _, num_doc_per_question = tf.unique_with_counts(support2question)
    offsets = tf.cumsum(num_doc_per_question, exclusive=True)
    doc_idx_for_support = tf.range(tf.shape(support2question)[0]) - tf.gather(offsets, support2question)

    doc_idx, start_pointer = tf.cond(
        is_eval,
        lambda: segment_top_k(start_scores, support2question, topk)[:2],
        lambda: (tf.expand_dims(answer2support, 1), tf.expand_dims(correct_start, 1)))

    doc_idx_flat = tf.reshape(doc_idx, [-1])
    start_pointer = tf.reshape(start_pointer, [-1])

    start_state = tf.gather_nd(encoded_support, tf.stack([doc_idx_flat, start_pointer], 1))
    start_state.set_shape([None, size])

    encoded_support_gathered = tf.gather(encoded_support, doc_idx_flat)
    question_state = tf.gather(question_state, doc_idx_flat)
    if bilinear:
        hidden_end = tf.layers.dense(tf.concat([question_state, start_state], 1), size, name="hidden_end")
        end_scores = tf.einsum('ik,ijk->ij', hidden_end, encoded_support_gathered)
    else:
        end_input = tf.concat([tf.expand_dims(start_state, 1) * encoded_support_gathered,
                               tf.gather(static_input, doc_idx_flat)], 2)

        hidden_end = tf.layers.dense(tf.concat([question_state, start_state], 1), size,
                                     name="hidden_end_1")
        hidden_end = tf.layers.dense(
            end_input, size, use_bias=False, name="hidden_end_2") + tf.expand_dims(hidden_end, 1)

        end_scores = tf.layers.dense(tf.nn.relu(hidden_end), 1, use_bias=False, name="end_scores")
        end_scores = tf.squeeze(end_scores, [2])

    end_scores = end_scores + tf.gather(support_mask, doc_idx_flat)

    def train():
        predicted_end_pointer = tf.argmax(end_scores, axis=1, output_type=tf.int32)
        return start_scores, end_scores, doc_idx, start_pointer, predicted_end_pointer

    def eval():
        # [num_questions * topk, support_length]
        left_mask = misc.mask_for_lengths(tf.cast(start_pointer, tf.int32),
                                          max_support_length, mask_right=False)
        right_mask = misc.mask_for_lengths(tf.cast(start_pointer + max_span_size, tf.int32),
                                           max_support_length)
        masked_end_scores = end_scores + left_mask + right_mask
        predicted_ends = tf.argmax(masked_end_scores, axis=1, output_type=tf.int32)

        return (start_scores, masked_end_scores,
                tf.gather(doc_idx_for_support, doc_idx_flat), start_pointer, predicted_ends)

    return tf.cond(is_eval, eval, train)


def bidaf_answer_layer(encoded_support_start, encoded_support_end, support_length,
                       support2question, answer2support, is_eval, topk=1, max_span_size=10000):
    # BiLSTM(M) = M^2 = encoded_support_end
    start_scores = tf.squeeze(tf.layers.dense(encoded_support_start, 1, use_bias=False), 2)
    end_scores = tf.squeeze(tf.layers.dense(encoded_support_end, 1, use_bias=False), 2)
    # mask out-of-bounds slots by adding -1000
    support_mask = misc.mask_for_lengths(support_length)
    start_scores = start_scores + support_mask
    end_scores = end_scores + support_mask
    return compute_spans(start_scores, end_scores, answer2support, is_eval,
                         support2question, topk=topk, max_span_size=max_span_size)


def san_answer_layer(size, encoded_question, question_length, encoded_support, support_length,
                     support2question, answer2support, is_eval, topk=1, max_span_size=10000,
                     num_steps=5, dropout=0.4, **kwargs):
    question_state = compute_question_state(encoded_question, question_length)
    question_state = tf.layers.dense(question_state, encoded_support.get_shape()[-1].value, tf.tanh)
    question_state = tf.gather(question_state, support2question)

    cell = tf.contrib.rnn.GRUBlockCell(size)

    all_start_scores = []
    all_end_scores = []

    support_mask = misc.mask_for_lengths(support_length)

    for i in range(num_steps):
        with tf.variable_scope('SAN', reuse=i > 0):
            question_state = tf.expand_dims(question_state, 1)
            support_attn = attention.bilinear_attention(
                question_state, encoded_support, support_length, False, False)[2]
            question_state = tf.squeeze(question_state, 1)
            support_attn = tf.squeeze(support_attn, 1)
            question_state = cell(support_attn, question_state)[0]

            hidden_start = tf.layers.dense(question_state, size, name="hidden_start")

            start_scores = tf.einsum('ik,ijk->ij', hidden_start, encoded_support)
            start_scores = start_scores + support_mask

            start_probs = segment_softmax(start_scores, support2question)
            start_states = tf.einsum('ij,ijk->ik', start_probs, encoded_support)
            start_states = tf.unsorted_segment_sum(start_states, support2question, tf.shape(question_length)[0])
            start_states = tf.gather(start_states, support2question)

            hidden_end = tf.layers.dense(tf.concat([question_state, start_states], 1), size, name="hidden_end")

            end_scores = tf.einsum('ik,ijk->ij', hidden_end, encoded_support)
            end_scores = end_scores + support_mask
            all_start_scores.append(start_scores)
            all_end_scores.append(end_scores)

    all_start_scores = tf.stack(all_start_scores)
    all_end_scores = tf.stack(all_end_scores)
    dropout_mask = tf.nn.dropout(tf.ones([num_steps, 1, 1]), 1.0 - dropout)
    all_start_scores = tf.cond(is_eval, lambda: all_start_scores * dropout_mask, lambda: all_start_scores)
    all_end_scores = tf.cond(is_eval, lambda: all_end_scores * dropout_mask, lambda: all_end_scores)

    start_scores = tf.reduce_mean(all_start_scores, axis=0)
    end_scores = tf.reduce_mean(all_end_scores, axis=0)

    return compute_spans(start_scores, end_scores, answer2support, is_eval, support2question, topk=topk,
                         max_span_size=max_span_size)



"""Credits: https://github.com/kefirski/pytorch_Highway"""

import torch.nn as nn
import torch.nn.functional as F


class Highway(nn.Module):
    def __init__(self, size, num_layers, f=F.tanh):
        super(Highway, self).__init__()

        self.num_layers = num_layers

        self.nonlinear = nn.ModuleList([nn.Linear(size, size) for _ in range(num_layers)])

        self.linear = nn.ModuleList([nn.Linear(size, size) for _ in range(num_layers)])

        self.gate = nn.ModuleList([nn.Linear(size, size) for _ in range(num_layers)])

        self.f = f

    def forward(self, x):
        """
            :param x: tensor with shape of [batch_size, size]
            :return: tensor with shape of [batch_size, size]
            applies σ(x) ⨀ (f(G(x))) + (1 - σ(x)) ⨀ (Q(x)) transformation | G and Q is affine transformation,
            f is non-linear transformation, σ(x) is affine transformation with sigmoid non-linearition
            and ⨀ is element-wise multiplication
            """

        for layer in range(self.num_layers):
            gate = F.sigmoid(self.gate[layer](x))

            nonlinear = self.f(self.nonlinear[layer](x))
            linear = self.linear[layer](x)

            x = gate * nonlinear + (1 - gate) * linear

        return x



# -*- coding: utf-8 -*-

import re
from typing import Mapping, List, Any, Union, Tuple, Optional

import numpy as np
import spacy
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import pairwise_distances

from jack.util.vocab import Vocab


def fill_vocab(qa_settings, vocab=None, lowercase=False, lemmatize=False, spacy_nlp=False):
    vocab = vocab or Vocab(unk=None)
    assert not vocab._frozen, 'Filling frozen vocabs does not make a lot fo sense...'
    for qa_setting in qa_settings:
        nlp_preprocess(qa_setting.question, vocab, lowercase, lemmatize, use_spacy=spacy_nlp)
        for s in qa_setting.support:
            nlp_preprocess(s, vocab, lowercase, lemmatize, use_spacy=spacy_nlp)
    return vocab


__pattern = re.compile('\w+|[^\w\s]')


def tokenize(text, pattern=__pattern):
    return __pattern.findall(text)


def token_to_char_offsets(text, tokenized_text):
    offsets = []
    offset = 0
    for t in tokenized_text:
        offset = text.index(t, offset)
        offsets.append(offset)
        offset += len(t)
    return offsets


def nlp_preprocess_all(qa_settings,
                       vocab: Vocab,
                       lowercase: bool = False,
                       lemmatize: bool = False,
                       with_lemmas: bool = False,
                       with_tokens_offsets: bool = False,
                       use_spacy: bool = False):
    assert not vocab._frozen, 'Filling frozen vocabs does not make a lot fo sense...'
    processed_questions = []
    processed_support = []
    for qa_setting in qa_settings:
        processed_questions.append(
            nlp_preprocess(qa_setting.question, vocab, lowercase, lemmatize, use_spacy=use_spacy))
        processed_support.append([])
        for s in qa_settings.support:
            processed_support[-1].append(nlp_preprocess(s, vocab, lowercase, lemmatize, use_spacy=use_spacy))
    return processed_questions, processed_support


__spacy_nlp = None


def spacy_nlp(parser=False, entity=False, matcher=False):
    import spacy
    global __spacy_nlp
    if __spacy_nlp is None:
        __spacy_nlp = spacy.load("en", parser=parser, entity=entity, matcher=matcher)
    return __spacy_nlp


def nlp_preprocess(text: str,
                   vocab: Vocab,
                   lowercase: bool = False,
                   lemmatize: bool = False,
                   with_lemmas: bool = False,
                   with_tokens_offsets: bool = False,
                   use_spacy: bool = False) \
        -> Tuple[List[str], List[int], int, Optional[List[str]],
                 Optional[List[int]]]:
    """Preprocesses a question and support:
    The steps include tokenization, lower-casing. It also includes the computation of token-to-character offsets for
    the support. Lemmatization is supported in 2 ways. If lemmatize is True then the returned tokens are lemmatized
    and the ids correspond to the lemma ids in the vocab. If with_lemmas and not lemmatize then an additional list
    of the lemmatized token in string form is returned.

    Returns:
        tokens, ids, length, lemmas or None, token_offsets or None
    """
    assert not with_lemmas or use_spacy, "enable spacy when using lemmas"
    assert not lemmatize or use_spacy, "enable spacy when using lemmas"

    if use_spacy:
        import spacy
        nlp = spacy_nlp()
        thistokenize = lambda t: nlp(t)
    else:
        thistokenize = tokenize
    if lowercase:
        text = text.lower()
    tokens = thistokenize(text)

    token_offsets = None
    lemmas = None
    if use_spacy:
        if with_lemmas:
            lemmas = [t.lemma_ for t in tokens]
        if with_tokens_offsets:
            token_offsets = [t.idx for t in tokens]
        tokens = [t.lemma for t in tokens] if lemmatize else [t.orth_ for t in tokens]
    else:
        # char to token offsets
        if with_tokens_offsets:
            token_offsets = token_to_char_offsets(text, tokens)

    length = len(tokens)
    ids = vocab(tokens)

    return tokens, ids, length, lemmas, token_offsets


def transpose_dict_of_lists(dict_of_lists: Mapping[str, list], keys: List[str]) \
        -> List[Mapping[str, Any]]:
    """Takes a dict of lists, and turns it into a list of dicts."""

    return [{key: dict_of_lists[key][i] for key in keys}
            for i in range(len(dict_of_lists[keys[0]]))]


def char_vocab_from_vocab(vocab):
    char_vocab = dict()
    char_vocab["PAD"] = 0
    for i in range(len(vocab)):
        w = vocab.get_sym(i)
        if w is not None:
            for c in w:
                if c not in char_vocab:
                    char_vocab[c] = len(char_vocab)
    return char_vocab


def stack_and_pad(values: List[Union[np.ndarray, int, float]], pad=0) -> np.ndarray:
    """Pads a list of numpy arrays so that they have equal dimensions, then stacks them."""
    if isinstance(values[0], int) or isinstance(values[0], float):
        return np.array(values)

    dims = len(values[0].shape)
    max_shape = [max(sizes) for sizes in zip(*[v.shape for v in values])]

    padded_values = []

    for value in values:
        pad_width = [(0, max_shape[i] - value.shape[i])
                     for i in range(dims)]
        padded_value = np.lib.pad(value, pad_width, mode='constant',
                                  constant_values=pad)
        padded_values.append(padded_value)

    return np.stack(padded_values)


def unique_words_with_chars(tokens, char_vocab, char_limit=20):
    vocab = dict()
    rev_vocab = list()
    unique_words = list()
    unique_word_lengths = list()
    token2unique = list()

    for j in range(len(tokens)):
        t2u = list()
        for w in tokens[j]:
            if w not in vocab:
                unique_word_lengths.append(min(char_limit, len(w)))
                unique_words.append([char_vocab.get(c, 0) for c in w[:char_limit]])
                vocab[w] = len(vocab)
                rev_vocab.append(w)
            t2u.append(vocab[w])
        token2unique.append(t2u)

    return unique_words, unique_word_lengths, token2unique, vocab, rev_vocab


def sort_by_tfidf(reference, candidates):
    tfidf = TfidfVectorizer(strip_accents="unicode", stop_words=spacy.en.STOP_WORDS, decode_error='replace')
    try:
        para_features = tfidf.fit_transform(candidates)
        q_features = tfidf.transform([reference])
    except ValueError:
        return [(i, 0.0) for i in range(len(candidates))]

    dists = pairwise_distances(q_features, para_features, "cosine").ravel()
    sorted_ix = np.lexsort((candidates, dists))  # in case of ties, use the earlier paragraph

    return [(i, 1.0 - dists[i]) for i in sorted_ix]



# -*- coding: utf-8 -*-
import os
import pickle
import zipfile

import yaml

from jack.io.embeddings.fasttext import load_fasttext
from jack.io.embeddings.glove import load_glove
from jack.io.embeddings.word_to_vec import load_word2vec


class Embeddings:
    """Wraps Vocabulary and embedding matrix to do lookups"""

    def __init__(self, vocabulary: dict, lookup, filename: str = None, emb_format: str = None):
        """
        Args:
            vocabulary:
            lookup:
            filename:
        """
        self.filename = filename
        self.vocabulary = vocabulary
        self.lookup = lookup
        self.emb_format = emb_format

    def get(self, word, default=None):
        _id = None
        if self.vocabulary is not None:
            _id = self.vocabulary.get(word, None)
        # Handling OOV words - Note: lookup[None] would return entire lookup table
        return self.lookup[_id] if _id is not None else default

    def __call__(self, word):
        return self.get(word)

    @property
    def shape(self):
        return self.lookup.shape

    def store(self, path):
        if not os.path.exists(path):
            os.mkdir(path)
        if self.filename is None:
            self.filename = os.path.join(path, "emb.pkl")
            self.emb_format = 'pkl'
            with open(self.filename, "wb") as f:
                pickle.dump(self, f)
        conf_file = os.path.join(path, "config.yaml")
        with open(conf_file, "w") as f:
            yaml.dump({"embedding_file": self.filename, "emb_format": self.emb_format}, f)

    @staticmethod
    def from_config(conf_file):
        with open(conf_file, "r") as f:
            config = yaml.load(f)
        if config["embedding_file"] is not None:
            return load_embeddings(config["embedding_file"], typ=config.get("emb_format", None))

    @staticmethod
    def from_dir(dir):
        with open(os.path.join(dir, "config.yaml"), "r") as f:
            config = yaml.load(f)
        if config["embedding_file"] is not None:
            return load_embeddings(config["embedding_file"], typ=config.get("emb_format", None))


def load_embeddings(file, typ='glove', **options):
    """
    Loads either GloVe or word2vec embeddings and wraps it into Embeddings

    Args:
        file: string, path to a file like "GoogleNews-vectors-negative300.bin.gz" or "glove.42B.300d.zip"
        typ: string, either "word2vec", "glove", "fasttext", "mem_map" or "pkl"
        options: dict, other options.
    Returns:
        Embeddings object, wrapper class around Vocabulary embedding matrix.
    """
    type_set = {"word2vec", "glove", "fasttext", "memory_map_dir", "pkl"}
    assert typ.lower() in type_set, "so far only {} foreseen".format(', '.join(type_set))

    if typ.lower() == "word2vec":
        return Embeddings(*load_word2vec(file, **options))

    elif typ.lower() == "glove":
        if file.endswith('.txt'):
            with open(file, 'rb') as f:
                return Embeddings(*load_glove(f), filename=file, emb_format=typ)
        elif file.endswith('.zip'):
            with zipfile.ZipFile(file) as zf:
                txtfile = file.split('/')[-1][:-4] + '.txt'
                with zf.open(txtfile, 'r') as f:
                    return Embeddings(*load_glove(f), filename=file, emb_format=typ)
        else:
            raise NotImplementedError

    elif typ.lower() == "fasttext":
        with open(file, 'rb') as f:
            return Embeddings(*load_fasttext(f), filename=file, emb_format=typ)

    elif typ.lower() == "memory_map_dir":
        from jack.io.embeddings.memory_map import load_memory_map_dir
        return load_memory_map_dir(file)

    elif typ.lower() == "pkl":
        with open(file, 'rb') as f:
            return pickle.load(f)

    else:
        raise ValueError("Unknown type: {}".format(type))



"""
This files merges two data files, both in JTR format, into a single JTR data file.
It assumes that the structure of instances is identical for both input files
and only concatenates the two instances lists.
It also assumes that the global variables are identical in both input files.
"""

import json
import sys


def main():

    if len(sys.argv) != 4:
        print('Wrong arguments for merging two data files in Jack format into one. Usage:')
        print('\tpython3 merge_JTR_data_files.py input1.json input2.json output.json')
    else:
        # load input 1
        with open(sys.argv[1], 'r') as inputfile1:
            content1 = json.load(inputfile1)

        # load input 2
        with open(sys.argv[2], 'r') as inputfile2:
            content2 = json.load(inputfile2)

        # define new 'meta' field
        meta_ = "Merged Content of {} and {}".format(content1['meta'], content2['meta'])

        # define new 'globals' field. Note: so far assuming same globals in both input files.
        assert (content1['globals']) == content2['globals']
        globals_ = content1['globals']

        # concatenating instances of both input files
        instances_ = content1['instances'] + content2['instances']

        # defining the dictionary for dumping into json
        merged_content = {'meta': meta_, 'globals': globals_, 'instances': instances_}

        # sanity check: nothing unexpected got lost or added
        assert len(content1['instances']) + len(content2['instances']) == len(merged_content['instances'])

        # summary print
        print('Merged file {} with {} into {}'.format(sys.argv[1],sys.argv[2],sys.argv[3]))
        print('Number of instances: input1: {} input2: {} output: {}'\
            .format(len(content1['instances']), len(content2['instances']), len(merged_content['instances'])))

        # dump merged content into JTR output file.
        with open(sys.argv[3], 'w') as outputfile:
            json.dump(merged_content, outputfile)



if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import tensorflow as tf


def fixed_dropout(xs, keep_prob, noise_shape, seed=None):
    """
    Apply dropout with same mask over all inputs
    Args:
        xs: list of tensors
        keep_prob:
        noise_shape:
        seed:

    Returns:
        list of dropped inputs
    """
    with tf.name_scope("dropout", values=xs):
        noise_shape = noise_shape
        # uniform [keep_prob, 1.0 + keep_prob)
        random_tensor = keep_prob
        random_tensor += tf.random_uniform(noise_shape, seed=seed, dtype=xs[0].dtype)
        # 0. if [keep_prob, 1.0) and 1. if [1.0, 1.0 + keep_prob)
        binary_tensor = tf.floor(random_tensor)
        outputs = []
        for x in xs:
            ret = tf.div(x, keep_prob) * binary_tensor
            ret.set_shape(x.get_shape())
            outputs.append(ret)
        return outputs
